"""Tests for the prevo package.

(VERY partial)
"""

# Standard library
from pathlib import Path

# Non standard
import pytest

# local imports
import prevo
from prevo.measurements import SavedCsvData


DATAFOLDER = Path(prevo.__file__).parent / '..' / 'data/manip'

NAMES = 'P', 'T', 'B1'

FILENAMES = {
    'P': 'Vacuum_Pressure.tsv',
    'T': 'Vacuum_Temperature.tsv',
    'B1': 'Vacuum_SourceBath.tsv',
}

MEAS_NUMBERS = {'P': 22553, 'T': 11271, 'B1': 9883}

n = 8  # measurement line number to check when loading full file
# (considering first data line is numbered 1)

nrange = (5, 10)  # range for partial loading test
nred = n - nrange[0] + 1  # corresponding line in the partial data

LINES = {
    'P': (1616490450.506, 0.167, 2727.25),
    'T': (1616490515.961, 2.623, 26.8932, 25.4829),
    'B1': (1616490514.585, 0.091, 24.2640),
}

measP = SavedCsvData('P', DATAFOLDER / FILENAMES['P'])
measT = SavedCsvData('T', path=DATAFOLDER / FILENAMES['T'])
measB1 = SavedCsvData(
    name='B1',
    path=DATAFOLDER / FILENAMES['B1'],
    csv_separator='\t',
)

SAVED_DATA = {'P': measP, 'T': measT, 'B1': measB1}


def test_number_of_measurements():  # test with a variety of call options
    ns = [sdata.number_of_measurements() for sdata in SAVED_DATA.values()]
    assert ns == list(MEAS_NUMBERS.values())


@pytest.mark.parametrize('name', NAMES)
def test_load_full(name):             # test full loading of data from file
    sdata = SAVED_DATA[name]
    sdata.load()
    assert len(sdata.data) == MEAS_NUMBERS[name]
    assert tuple(sdata.data.loc[n - 1].round(decimals=4)) == LINES[name]


@pytest.mark.parametrize('name', NAMES)
def test_load_partial(name):
    sdata = SAVED_DATA[name]
    sdata.load(nrange)  # test partial loading of data
    assert len(sdata.data) == nrange[1] - nrange[0] + 1
    assert tuple(sdata.data.loc[nred - 1].round(decimals=4)) == LINES[name]



"""Programs to make automatic temporal patterns of settings."""

# ----------------------------- License information --------------------------

# This file is part of the prevo python package.
# Copyright (C) 2022 Olivier Vincent

# The prevo package is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# The prevo package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with the prevo python package.
# If not, see <https://www.gnu.org/licenses/>

# Standard library imports
from datetime import timedelta
from threading import Thread, Event
from pathlib import Path
import json

# Other modules
import oclock
import matplotlib.pyplot as plt


# ================================== Config ==================================


time_factors = {'s': 1, 'min': 60, 'h': 3600}  # used for slope calculations


# ============================= Misc. Functions ==============================

def _get_and_check_input(entry, possible_inputs):
    try:
        (quantity, values), = entry.items()
    except ValueError:
        msg = f"Only one input quantity allowed, not {list(entry.keys())}"
        raise ValueError(msg)

    if possible_inputs is not None and quantity not in possible_inputs:
        msg = f"Settings must only have one of {possible_inputs} as key."
        raise ValueError(msg)
    else:
        return quantity, values


def _format_time(t):
    """format hh:mm:ss str time into timedelta if not already timedelta."""
    try:
        dt = t.total_seconds()
    except AttributeError:
        dt = oclock.parse_time(t).total_seconds()
    finally:
        return dt


def _seconds_to_hms(t):
    """convert seconds to hours, minutes, seconds tuple."""
    h = t // 3600
    ss = t % 3600
    m = ss // 60
    s = ss % 60
    return h, m, s


def _circular_permutation(x):
    return x[1:] + [x[0]]


# ============= Classes to make and manage temperature programs ==============


class Program:
    """Class for managing programmable cycles for control of devices."""

    def __init__(
        self,
        control=None,
        repeat=0,
        **steps,
    ):
        """Initiate temperature cycle(s) program on device.

        Parameters
        ----------
        - control: object of the Control class (or child classes)
                   (can be added later after instantiation)
                   # Note: Program will inherit possible_inputs from the
                           Control object either upon instantiation or
                           when it is defined later.
        - repeat: number of times the cycle is repeated (int); if 0 (default),
                  the program is done only once.
        - steps: kwargs/dict with the following keys:
            * durations: list tt of step durations (timedelta or str 'h:m:s')
            * p=, rh=, aw=, T=: list vv of step target values

        Notes
        -----
        - tt[i] is the duration between steps vv[i] and vv[i + 1].
        - When a cycle is done, the last step loops back to the first step,
            i.e. tt[-1] connects vv[-1] and vv[0]

        Example
        -------
        pp = [3100, 3100, 850]
        tt = ['3::', '1:30:', '::']

        ctrl = Control()  # see Control subclasses for details on instantiation

        # EITHER:
        program = ctrl.program(durations=tt, p=pp, repeat=2)
        # OR:
        program = Program(ctrl, durations=tt, p=pp, repeat=2)

        creates the following program (each dot represents 30 min):

        p (Pa) 3000 ......         ......        ......       .
                          .       |      .      |      .      |
                           .      |       .     |       .     |
               850          .......        ......        ......

        program.plot()   # check everything is ok
        program.run()    # start program
        program.running  # check whether program is running or not
        program.stop()   # stop program
        """
        self.control = control
        self.repeat = repeat
        self.durations = steps.pop('durations')  # list of step durations

        # quantity: 'p', 'rh', 'T', etc., origins: initial values for each ramp
        self.quantity, self.origins = _get_and_check_input(
            steps,
            possible_inputs=self.possible_inputs
        )
        self.targets = _circular_permutation(self.origins)  # loops back to beginning

        self.stop_event = Event()
        self.stop_event.set()

    def __repr__(self):
        msg = f'{self.__class__} with {len(self.durations)} steps of ' \
              f'{self.quantity.upper()} and {self.repeat} repeats.'
        return msg

    def _run(self):
        """Start program in a blocking manner, stop if stop_event is set."""

        if self.control is None:
            msg = 'Control object that the program acts upon not defined yet. '
            msg += 'Please defined a `program.control` attribute of Control type.'
            raise ValueError(msg)

        if self.running:
            msg = 'Program already running. No action taken.'
            self.control._manage_message(msg, force_print=True)
            return

        self.stop_event.clear()

        for n in range(self.repeat + 1):

            msg = f'------ PROGRAM ({self.quantity})--- NEW CYCLE {n + 1} / {self.repeat + 1}'
            self.control._manage_message(msg, force_print=True)

            for v1, v2, duration in zip(self.origins, self.targets, self.durations):
                self.control._ramp(duration, **{self.quantity: (v1, v2)})
                if self.stop_event.is_set():
                    msg = f'------ PROGRAM ({self.quantity})--- STOPPED'
                    self.control._manage_message(msg, force_print=True)
                    return
        else:
            msg = f'------ PROGRAM ({self.quantity})--- FINISHED'
            self.control._manage_message(msg, force_print=True)

    def run(self):
        """Start program in a non-blocking manner."""
        Thread(target=self._run).start()

    def stop(self):
        """Interrupt program."""
        self.stop_event.set()
        self.control.stop()

    def plot(self, time_unit='min'):
        """Same input as cycle(), to visualize the program before running it.

        Note: time_unit can be in 'h', 'min', 's'
        """
        fig, ax = plt.subplots()
        ax.grid()

        t = 0
        for v1, v2, duration in zip(self.origins, self.targets, self.durations):
            dt = _format_time(duration) / time_factors[time_unit]  # h, min, s
            ax.plot([t, t + dt], [v1, v2], '-ok')
            t += dt

        ax.set_xlabel(f'time ({time_unit})')
        ax.set_ylabel(f'{self.quantity}')

        fig.show()

    def copy(self):
        """Return program with same characteristics of current one.

        can be useful to run the same program on different sensors/devices.
        """
        return Program(**self.info)

    def save(self, savepath='.', filename=None):
        """Save program to JSON file. Use load() to regenerate program.

        If filename is None, use default filename.
        """
        filename =  self.default_filename if filename is None else filename
        file = Path(savepath) / filename
        with open(file, 'w', encoding='utf8') as f:
            json.dump(self.info, f, indent=4, ensure_ascii=False)

    @classmethod
    def load(cls, savepath='.', filename=None):
        """Load program from file (filename NEEDS to be supplied here)."""
        if filename is None:
            raise ValueError('Filename needs to be provided')
        file = Path(savepath) / filename
        with open(file, 'r', encoding='utf8') as f:
            info = json.load(f)
        return cls(**info)

    @staticmethod
    def _to_json(file, data):
        """Load python data (dict or list) from json file"""

    @property
    def info(self):
        """Gathers useful info to regenerate the program in a dict.

        Note: self.control not in there because not easily represented by
        a string / number.
        """
        info_dict = {
            'repeat': self.repeat,
            'durations': self.durations,
            self.quantity: self.origins,
        }
        return info_dict

    @property
    def default_filename(self):
        return f'Program_{self.quantity}.json'

    @property
    def cycle_duration(self):
        """Duration of a single cycle of the program."""
        dt = 0
        for duration in self.durations:
            dt += _format_time(duration)  # in seconds
        return timedelta(seconds=dt)

    @property
    def total_duration(self):
        """Duration of a all cycles including repeats."""
        return (self.repeat + 1) * self.cycle_duration

    @property
    def running(self):
        is_running = False if self.stop_event.is_set() else True
        return is_running

    @property
    def control(self):
        return self._control

    @control.setter
    def control(self, value):
        self._control = value
        if self._control is None:
            self.possible_inputs = None
        else:
            self.possible_inputs = self._control.possible_inputs


class Stairs(Program):
    """Special program consisting in plateaus of temperature, of same duration."""

    def __init__(self,
                 control=None,
                 duration=None,
                 repeat=0,
                 **steps):
        """Specific program with a succession of constant setting plateaus.

        Parameters
        ----------
        - control: object of the Control class (or child classes)
                   (can be added later after instantiation)
                   # Note: Program will inherit possible inputs from the
                           Control object either upon instantiation or
                           when it is defined later.
        - duration: (str of type 'hh:mm:ss' or timedelta): if not None, sets
          the duration of every plateau to be the same. If None, a list of
          durations must be supplied within **steps (key 'durations').

        - repeat: number of times the cycle is repeated (int); if 0 (default),
                  the program is done only once.

        - **steps:
            - do not correspond to points in the program, but segments.
            - if duration is None: must contain the key 'durations' with a list
              (tt) of plateau durations as value
            - if duration is not None: 'durations' key is ignored
            - other key must be p=, rh= etc. with the list of plateau setpoints
              (vv) s values.

        Notes
        -----
        - tt[i] is the duration of step i at setting vv[i] (this is different
          from the behavior of Control.program())
        - similarly to Control.program(), the program loops back to the
          beginning, i.e. the last setting sent to the bath is the first one
          in the list.

        Example 1
        ---------
        rh = [90, 70, 50, 30]

        ctrl = Control()  # see Control subclasses for details on instantiation

        # EITHER:
        program = ctrl.stairs(duration='1::', rh=rh, repeat=1)
        # OR:
        program = Stairs(ctrl, duration='1::', rh=rh, repeat=1)

        creates the following program (each dot represents 15 min):

        %RH  90 ....            ....           .
                   |           |   |           |
             70     ....       |    ....       |
                       |       |       |       |
             50         ....   |        ....   |
                           |   |           |   |
             30             ....            ....

        Methods to run, check and stop the program are the same as for the
        Program class:
            program.plot()
            program.run()
            program.running
            program.stop()

        Example 2
        ---------
        pp = [3170, 2338]
        tt = ['1::', '3::']
        Control().stairs(durations=tt, rh=rh, repeat=2)

        Does the following when run (each dot represents 15 minutes):

        p (Pa)  3170 ....            ....            ....           .
                        |           |   |           |   |           |
                        |           |   |           |   |           |
                2338     ............    ............    ............
        """
        self.control = control  # To be able to get possible_inputs automatically
        if duration is None:
            durations = steps.pop('durations')  # list of step durations

        qty, step_plateaus = _get_and_check_input(
            steps,
            possible_inputs=self.possible_inputs
        )
        step_points = sum([[v, v] for v in step_plateaus], [])

        if duration is None:
            step_durations = sum([[dt, '::'] for dt in durations], [])
        else:
            step_durations = [duration, '::'] * len(step_plateaus)

        formatted_steps = {'durations': step_durations, qty: step_points}

        super().__init__(control=control,
                         repeat=repeat,
                         **formatted_steps)


class Teeth(Program):
    """Plateaus of fixed duration separated by ramps of constant slope."""

    def __init__(
        self,
        control=None,
        slope=None,
        slope_unit='/min',
        plateau_duration='::',
        start='plateau',
        repeat=0,
        **steps,
    ):
        """Plateaus of fixed duration separated by ramps of constant slope.

        Parameters
        ----------
        - control: object of the Control class (or child classes)
                   (can be added later after instantiation)
                   # Note: Program will inherit possible inputs from the
                           Control object either upon instantiation or
                           when it is defined later.
        - slope: rate of change of the parameter specified in **steps. For
          example, if steps contains rh=..., the slope is in %RH / min, except
          if a different time unit is specified in slope_unit.

        - slope_unit: can be '/s', '/min', '/h'

        - plateau_duration : duration of every plateau separating the ramps
          (as as hh:mm:ss str or as a timedelta object)

        - start: can be 'plateau' (default, start with a plateau) or 'ramp'
          (start directly with a ramp).

        - repeat: number of times the cycle is repeated (int); if 0 (default),
                  the program is done only once.

        - **steps:
            - correspond to the values of the plateaus.
            - only key must be p=, rh= etc. with the list of plateau setpoints
              as values.

        Example 1
        ---------
        rh = [90, 70, 90, 50]

        ctrl = Control()  # see Control subclasses for details on instantiation

        # EITHER:
        program = ctrl.teeth(40, '/h', '1::', rh=rh, repeat=1)
        # OR:
        program = Teeth(ctrl, 40, '/h', '1::', rh=rh, repeat=1)

        creates the following program (15 min / dot, ramps at 40%RH / hour):

        %RH  90 ....      ....          ....     ....          .
                    .    .    .        .    .        .        .
             70      ....      .      .      ....     .      .
                                .    .                 .    .
             50                  ....                   ....

        Methods to run, check and stop the program are the same as for the
        Program class:
            program.plot()
            program.run()
            program.running
            program.stop()

        Example 2
        ---------
        pp = [3000, 2000, 3000, 1000]
        Control().teeth(25, '/min', '1:20:', p=pp, start='ramp', repeat=1)

        Does the following when run (20 min / dot, ramps at 25 Pa / min):

        p (Pa)  3000 .      ....          ....     ....          ....
                      .    .    .        .    .        .        .
                2000   ....      .      .      ....     .      .
                                  .    .                 .    .
                1000               ....                   ....
        """
        self.control = control  # To be able to get possible_inputs automatically
        self.slope = slope
        self.slope_unit = slope_unit

        qty, values = _get_and_check_input(
            steps,
            possible_inputs=self.possible_inputs
        )
        next_values = _circular_permutation(values)

        dt_ramps = [self._slope_to_time(vals) for vals in zip(values, next_values)]

        dts = sum([[plateau_duration, dt] for dt in dt_ramps], [])
        pts = sum([[v, v] for v in values], [])

        if start == 'plateau':
            step_durations = dts
            step_points = pts
        elif start == 'ramp':
            step_durations = _circular_permutation(dts)
            step_points = _circular_permutation(pts)
        else:
            msg = f"start parameter must be 'plateau' or 'ramp', not {start}"
            raise ValueError(msg)

        formatted_steps = {'durations': step_durations, qty: step_points}

        super().__init__(control,
                         repeat=repeat,
                         **formatted_steps)

    def _slope_to_time(self, values):
        """values is a tuple (v1, v2) of start and end values."""

        v1, v2 = values

        dvdt = self.slope / time_factors[self.slope_unit.strip('/')]  # in qty / second

        dt = abs((v2 - v1) / dvdt)  # ramp time in seconds
        h, m, s = _seconds_to_hms(dt)

        return timedelta(hours=h, minutes=m, seconds=s)



"""Record several sensors as a function of time with interactive CLI and graph."""

# ----------------------------- License information --------------------------

# This file is part of the prevo python package.
# Copyright (C) 2022 Olivier Vincent

# The prevo package is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# The prevo package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with the prevo python package.
# If not, see <https://www.gnu.org/licenses/>

# Standard library imports
from abc import ABC, abstractmethod
from datetime import datetime
from pathlib import Path
from threading import Event, Thread
from queue import Queue, Empty
from traceback import print_exc
import os
import time

# Non-standard imports
from tqdm import tqdm
import oclock
from clivo import CommandLineInterface, ControlledProperty, ControlledEvent

# Local imports
from ..control import RecordingControl
from ..misc import mode_to_names

# ================================ MISC Tools ================================


def try_func(function):
    """Decorator to make functions return & print errors if errors occur"""

    def wrapper(*args, **kwargs):
        try:
            function(*args, **kwargs)

        except Exception:
            try:
                nmax, _ = os.get_terminal_size()
            except OSError:
                nmax = 79
            print("\n")
            print("=" * nmax)
            # args is normally the object (self), will print its repr
            print(f"ERROR in {function.__name__}() for args={args}, kwargs={kwargs}")
            print_exc()
            print("=" * nmax)
            print("\n")
            return

    return wrapper


# ----------------------------------------------------------------------------
# =============================== SENSOR CLASS ===============================
# ----------------------------------------------------------------------------


class SensorError(Exception):
    pass


class SensorBase(ABC):
    """Abstract base class for sensor acquisition."""

    # If one wants to control specific properties of the sensor through
    # the CLI in Record or through pre-defined programs, it's possible
    # to specify them here, and they will be automatically added to
    # the controlled_properties attribute of the corresponding Recording
    # NOTE: this needs to be a class attribute and not an instance attribute,
    # because these properties are checked before the Sensor object is
    # instantiated.
    controlled_properties = ()

    def __init__(self):
        # (optional) : specific sensor errors to catch, can be an Exception
        # class or an iterable of exceptions; if not specified in subclass,
        # any exception is caught.
        self.exceptions = ()

    def __enter__(self):
        """Context manager for sensor (enter). Optional."""
        return self

    def __exit__(self, exc_type, exc_val, traceback):
        """Context manager for sensor (exit). Optional."""
        pass

    @property
    @abstractmethod
    def name(self):
        """Name of sensor, Must be a class attribute."""
        pass

    def _read(self):
        """Read sensor.

        Either this method or _get_data() must be defined in subclasses.
        """
        with oclock.measure_time() as data:
            data["values"] = self._get_data()
        return data

    def _get_data(self):
        """Use this instead of subclassing _read() when you want the time
        of measurement and uncertainty on time to be added automatically.
        (useful when sensor does not return time information)"""
        return ()

    def read(self):
        """Read sensor and throw SensorError if measurement fails."""
        try:
            data = self._read()
        except self.exceptions as e:
            raise SensorError(f"Impossible to read {self.name} sensor: {e}")
        else:
            return data


# ----------------------------------------------------------------------------
# ============================= RECORDING CLASS ==============================
# ----------------------------------------------------------------------------


timer_ppty = ControlledProperty(
    attribute="interval",
    readable="Δt (s)",
    commands=("dt",),
)

active_ppty = ControlledProperty(
    attribute="active",
    readable="Rec. ON",
    commands=("on",),
)

saving_ppty = ControlledProperty(
    attribute="saving",
    readable="Sav. ON",
    commands=("save",),
)


class RecordingBase(ABC):
    """Base class for recording sensor periodically.

    Can be used as is, or can be fed to the Record class for extra features
    (live graph, metadata saving, etc.)
    """

    # Warnings when queue size goes over some limits
    queue_warning_limits = 1e1, 1e2, 1e3, 1e4, 1e5, 1e6, 1e7, 1e8, 1e9

    def __init__(
        self,
        Sensor,
        path=".",
        dt=1,
        ctrl_ppties=(),
        active=True,
        saving=True,
        continuous=False,
        warnings=False,
        precise=False,
        immediate=True,
        programs=(),
        control_params=None,
        dt_save=1.3,
        dt_check=0.9,
    ):
        """Initialize a Recording object.

        Parameters
        ----------
        Sensor : subclass of SensorBase
            Sensor class to record data from.
        path : str, default="."
            Directory where data will be recorded.
        dt : float, default=1
            Time interval (in seconds) between sensor readings.
        ctrl_ppties : iterable, optional
            Iterable of ControlledProperty objects to control during recording,
            in addition to default properties (time interval and active on/off).
        active : bool, default=True
            If False, data reading is disabled until `self.active` is set to True.
        saving : bool, default=True
            If False, data will not be saved to file until `self.saving` is set
            to True. Data acquired while `saving=False` will not be saved later.
        continuous : bool, default=False
            If True, data is acquired as fast as possible from the sensor.
        warnings : bool, default=False
            If True, print warnings from the Timer (e.g., loop too short).
        precise : bool, default=False
            If True, use a precise timer in `oclock` (see `oclock.Timer`).
        immediate : bool, default=True
            If False, changes to the timer (e.g., interval) occur at the next
            timestep. If True, a new data point is taken immediately.
        programs : iterable, optional
            Iterable of Program objects (or subclasses) for predefined temporal
            patterns of property changes (e.g., activate/deactivate sensor at
            specific times).
        control_params : dict, optional
            Dictionary of `{command: kwargs}` to pass to program controls
            (e.g., `dt`, `range_limits`). If None, default parameters are used
            (see `RecordingControl` class).
        dt_save : float, default=1.3
            Time interval (in seconds) for writing queues to files.
        dt_check : float, default=0.9
            Time interval (in seconds) for checking queue sizes.
        """
        self.Sensor = Sensor
        self.name = Sensor.name
        self.timer = oclock.Timer(
            interval=dt,
            name=self.name,
            warnings=warnings,
            precise=precise,
        )
        self.immediate = immediate

        self.path = Path(path)
        self.path.mkdir(exist_ok=True)

        self.dt_save = dt_save
        self.dt_check = dt_check

        # Queues in which data is put (NEED to be defined before self.saving)
        self.queues = {
            "saving": Queue(),
            "plotting": Queue(),
        }
        # Events that need to be set to put data in each data queue
        self.queue_events = {
            "saving": Event(),
            "plotting": Event(),
        }

        # NOTE: DO NOT change to self.active = active because this can cause
        # problems upon init of the recording when self.sensor is not defined
        # yet, if the active setter uses properties of the sensor (e.g. to
        # turn it on or off)
        self._active = (
            active  # can be set to False to temporarily stop recording from sensor
        )

        # NOTE: Here, KEEP saving instead of _saving because one needs to
        # not only set the status but also activate the queues, etc.
        self.saving = saving  # can be set to False to temporarily not save data to file

        self.continuous = continuous

        # To be defined in subclass.
        # Object that manages how data is written to files.
        # Must have a .file attribute whith is a pathlib.Path object, and
        # .init_file() method to create / init the file when recording starts
        self.file_manager = None

        # Iterable of the recording properties that the program / CLI control.
        # Possibility to add other properties in subclasses
        # (need to be of type ControlledProperty or subclass)
        if self.continuous:
            # If sensor instantiated as continuous (e.g. streaming cameras)
            # we assume that the timer is internally controlled in the sensor
            # so that it does not make sense to control the time interval
            # from the CLI.
            # If one has a sensor when one wants to switch from continuous
            # to not continuous, it's better to instantiate it as not
            # continuous (possibly with on=False) and then switch to
            # continuous and on=True.
            default_ppties = (active_ppty, saving_ppty)
        else:
            default_ppties = (active_ppty, saving_ppty, timer_ppty)
        self.controlled_properties = default_ppties + ctrl_ppties
        self._add_sensor_controlled_properties()
        self._generate_ppty_dict()

        # Optional temporal programs to make controlled properties evolve.
        self._init_programs(
            programs=programs,
            control_params=control_params,
        )

        self.threads = []
        self.internal_stop = Event()

    def __repr__(self):
        return f"{self.__class__.__name__} ({self.name})"

    @property
    def active(self):
        """In case a bunch of things (e.g. sensor start/stop) need to be done
        when changing the active property --> subclass this"""
        return self._active

    @active.setter
    def active(self, value):
        self._active = value

    @property
    def saving(self):
        """Temporarily activate/disable saving of data to file"""
        return self._saving

    @saving.setter
    def saving(self, value):
        self._saving = value
        if self._saving:
            self.activate_queue("saving")
        else:
            self.deactivate_queue("saving")

    @property
    def interval(self):
        """Timer interval for the sampling of data from the sensor"""
        return self.timer.interval

    @interval.setter
    def interval(self, value):
        self.timer.set_interval(value, immediate=self.immediate)

    def activate_queue(self, kind):
        """Activate saving of data into chosen queue.

        kind can be 'saving', 'plotting', etc.
        """
        self.queue_events[kind].set()

    def deactivate_queue(self, kind):
        """Dectivate saving of data into chosen queue.

        kind can be 'saving', 'plotting', etc.
        """
        self.queue_events[kind].clear()

    # Private methods --------------------------------------------------------

    def _add_sensor_controlled_properties(self):
        for ppty in self.Sensor.controlled_properties:
            new_attribute = "sensor." + ppty.attribute
            new_ppty = ControlledProperty(
                attribute=new_attribute,
                readable=ppty.readable,
                commands=ppty.commands,
            )
            self.controlled_properties += (new_ppty,)

    def _generate_ppty_dict(self):
        """Associate property commands to property objects"""
        self.ppty_commands = {}
        for ppty in self.controlled_properties:
            for ppty_cmd in ppty.commands:
                self.ppty_commands[ppty_cmd] = ppty

    def _init_programs(self, programs, control_params):
        self.programs = ()
        for supplied_program in programs:
            # In case same program is supplied to different recordings,
            # because a program object cannot be run several times in parallel.
            program = supplied_program.copy()

            ppty_cmd = program.quantity
            ppty = self.ppty_commands[ppty_cmd]
            if control_params is not None:
                control_kwargs = control_params.get(ppty_cmd, {})
            else:
                control_kwargs = {}
            log_filename = f"Control_Log_{self.name}_[{ppty.readable}].txt"
            program.control = RecordingControl(
                recording=self,
                ppty=ppty,
                log_file=log_filename,
                savepath=self.path,
                **control_kwargs,
            )
            self.programs += (program,)

    def _stop_programs(self):
        for program in self.programs:
            program.stop()

    def _set_property(self, ppty, value):
        """Set property of recording.

        recording: RecordingBase object or subclass on which property is applied
        ppty: ControlledProperty object or subclass.
        value: value of property to apply.
        """
        if ppty not in self.controlled_properties:
            return
        try:
            # avoids having to pass a convert function
            exec(f"self.{ppty.attribute} = {value}")
        except Exception as e:
            print(
                f"WARNING: Could not set {ppty.readable} for {self.name} to {value}.\n Exception: {e}"
            )

    # Compulsory methods / properties to subclass ----------------------------

    @abstractmethod
    def save(self, measurement, file):
        """Write data of measurement to (already open) file.

        file is the file object yielded by the open() context manager.
        """
        pass

    # Optional methods to subclass -------------------------------------------

    def format_measurement(self, data):
        """How to format the data given by self.Sensor.read().

        Returns a measurement object (e.g. dict, value, custom class etc.)."""
        return data

    def after_measurement(self):
        """Define what to do after measurement has been done and formatted.

        Acts on the recording object but does not return anything.
        (Optional)
        """
        pass

    # General methods and attributes (can be subclassed if necessary) --------

    def print_info_on_failed_reading(self, status):
        """Displays relevant info when reading fails."""
        t_str = datetime.now().isoformat(sep=" ", timespec="seconds")
        if status == "failed":
            print(f"{self.name} reading failed ({t_str}). Retrying ...")
        elif status == "resumed":
            print(f"{self.name} reading resumed ({t_str}).")

    # ======================= Data reading from sensor =======================

    def _data_read(self):
        """Reading data when the sensor context manager is active.

        (see self.data_read())
        """
        failed_reading = False  # True temporarily if reading fails

        # Without this here, the first data points are irregularly spaced.
        self.timer.reset()

        while not self.internal_stop.is_set():
            if not self.active:
                if not self.continuous:
                    # to avoid checking too frequently if active or not.
                    self.timer.checkpt()
                continue

            try:
                data = self.sensor.read()

            # Measurement has failed .........................................
            except SensorError:
                print_exc()
                if not failed_reading:  # means it has not failed just before
                    self.print_info_on_failed_reading("failed")
                failed_reading = True

            # Measurement is OK ..............................................
            else:
                if failed_reading:  # means it has failed just before
                    self.print_info_on_failed_reading("resumed")
                    failed_reading = False

                measurement = self.format_measurement(data)
                self.after_measurement()

                for queue, event in zip(
                    self.queues.values(), self.queue_events.values()
                ):
                    if event.is_set():
                        queue.put(measurement)

            # Below, this means that one does not try to acquire data right
            # away after a fail, but one waits for the usual time interval
            finally:
                if not self.continuous:
                    self.timer.checkpt()

    @try_func
    def data_read(self):
        """Read data from sensor and store it in data queues."""

        # In order to avoid loosing a sensor at initialization, e.g. because
        # the resource is already talking to another program, we try to
        # instantiate the class several times
        number_of_trials = 10
        error = None
        for n in range(number_of_trials):
            # No need to continue trying if the program is stopped
            if not self.running:
                return
            try:
                self.sensor = self.Sensor()
            except Exception as e:
                error = e
                print(
                    f"Error trying to instantiate sensor [{self.Sensor.name}]\n"
                    f"(trial {n + 1}/{number_of_trials}). Retrying in 1s ..."
                )
                time.sleep(1)
            else:
                if n > 0:
                    print(f"Sensor [{self.Sensor.name}] finally instantiated!")
                break
        else:
            raise RuntimeError(
                f"Impossible to instantiate sensor [{self.Sensor.name}] "
                f"after {n + 1} trials, due to error:\n{error}"
            )

        with self.sensor:
            # Initial setting of properties is done here in case one of the
            # properties acts on the sensor object, which is not defined
            # before this point.
            self._apply_ppties()
            self._start_programs()
            self._data_read()

    # ========================== Write data to file ==========================

    def _try_save(self, measurement, file, attempts=3):
        """Try saving data. If not, ignore"""
        error = False
        for attempt in range(attempts):
            try:
                self.save(measurement, file)
            except Exception as e:
                error = True
                print(
                    f"Error saving {measurement} for {self.name}: {e}. "
                    f"Attempt {attempt + 1}/{attempts}"
                )
            else:
                if error:
                    print(
                        f"Success saving {measurement} for {self.name} "
                        f"at attempt {attempt + 1}"
                    )
                return
        else:
            print(
                f"Impossible saving {measurement} for {self.name}; "
                "will be missing from data"
            )

    @try_func
    def data_save(self):
        """Save data that is stored in a queue by data_read."""

        saving_queue = self.queues["saving"]
        saving_timer = oclock.Timer(interval=self.dt_save)

        self.file_manager.init_file()

        while not self.internal_stop.is_set():
            # Open and close file at each cycle to be able to save periodically
            # and for other users/programs to access the data simultaneously
            with open(self.file_manager.path, "a", encoding="utf8") as file:
                # Get all data from saving queue as long as the current
                # iteration of the timer is still active.
                # - If there is not a lot of data to save, the while loop will
                #   break immediately and go to the saving_timer.checkpt()
                #   waiting phase.
                # - If there is a lot of data to save and/or it takes too long
                #   then the saving_timer interval will be exceeded and the
                #   while loop will exit, thus forcing the saving file to be
                #   closed and re-opened.
                while not saving_timer.interval_exceeded:
                    try:
                        measurement = saving_queue.get(timeout=self.dt_save)
                    except Empty:
                        pass
                    else:
                        if measurement is not None:
                            self._try_save(measurement, file)

                    if self.internal_stop.is_set():  # Move to buffering waitbar
                        break

            # periodic check whether there is data to save
            # This is outside of the with statement in order to close the
            # file as soon as possible when not in use.
            saving_timer.checkpt()

        # Buffering waitbar --------------------------------------------------

        if not saving_queue.qsize():
            return

        print(f"Data buffer saving started for {self.name}")

        # The nested statements below, similarly to above, ensure that
        # self.file_manager.path is opened and closed regularly to avoid
        # loosing too much data if there is an error.

        with tqdm(total=saving_queue.qsize()) as pbar:
            while True:
                try:
                    with open(self.file_manager.path, "a", encoding="utf8") as file:
                        saving_timer.reset()
                        while not saving_timer.interval_exceeded:
                            measurement = saving_queue.get(timeout=self.dt_save)
                            self._try_save(measurement, file)
                            pbar.update()
                except Empty:
                    break

        print(f"Data buffer saving finished for {self.name}")

    # ========================== Check queue sizes ===========================

    def _check_queue_size(self, queue, q_size_over, q_type):
        """Check that queue does not go beyond specified limits"""
        for qmax in self.queue_warning_limits:
            if queue.qsize() > qmax:
                if not q_size_over[qmax]:
                    print(
                        f"\nWARNING: {q_type} buffer size for {self.name}"
                        f"over {int(qmax)} elements"
                    )
                    q_size_over[qmax] = True

            if queue.qsize() <= qmax:
                if q_size_over[qmax]:
                    print(
                        f"\n{q_type} buffer size now below {int(qmax)}for {self.name}"
                    )
                    q_size_over[qmax] = False

    @try_func
    def check_queue_sizes(self):
        """Periodically verify that queue sizes are not over limits"""

        # Init queue warnings ------------------------------------------------

        lims = self.queue_warning_limits

        self.q_size_over = {}
        for queue_name in self.queues:
            self.q_size_over[queue_name] = {limit: False for limit in lims}

        # Check periodically -------------------------------------------------

        while not self.internal_stop.is_set():
            for queue_name, queue in self.queues.items():
                # No need to check if queue is not active
                if self.queue_events[queue_name].is_set():
                    self._check_queue_size(
                        queue=queue,
                        q_size_over=self.q_size_over[queue_name],
                        q_type=queue_name,
                    )

            self.internal_stop.wait(self.dt_check)

    # ====================== Start / stop recording ==========================

    def _apply_ppties(self):
        # Note: (_set_property() does not do anything if the property
        # does not exist for the recording of interest)
        for ppty_cmd, value in self.ppty_kwargs.items():
            ppty = self.ppty_commands[ppty_cmd]
            self._set_property(ppty=ppty, value=value)

    def _start_programs(self):
        """Optional pre-defined program for change of recording properties"""
        for program in self.programs:
            program.run()

    def _start_threads(self):
        """Start threads to read, save data etc."""
        for func in self.data_read, self.data_save, self.check_queue_sizes:
            thread = Thread(target=func)
            thread.start()
            # This is to be able to join threads in other programs using
            # the recording (i.e. wait for recording to finish)
            self.threads.append(thread)

    def start(self, **ppty_kwargs):
        """Start recording (nonblocking)"""

        # To be able to restart the recording after calling stop()
        self.internal_stop.clear()
        self.ppty_kwargs = ppty_kwargs

        try:
            self._start_threads()
        # Typically I think there should not be exceptions here because the
        # threads are in try_func decorator, but it's safe to catch
        # anything that might go wrong.
        except (Exception, KeyboardInterrupt):
            print(f"Error for {self.name}. Stopping recording.")
            print_exc()
            self.stop()

    def pause(self):
        self.active = False

    def resume(self):
        self.active = True

    def stop(self):
        self.internal_stop.set()
        self.timer.stop()
        self._stop_programs()

    @property
    def running(self):
        return not self.internal_stop.is_set()


# ----------------------------------------------------------------------------
# =============================== RECORD CLASS ===============================
# ----------------------------------------------------------------------------


class Record:
    """Asynchronous recording of several Recordings objects.

    Recordings objects are of type RecordingBase or subclasses."""

    def __init__(
        self,
        recordings,
        path=".",
        on_start=(),
        dt_request=0.7,
    ):
        """Initialize the base class for recording data asynchronously.

        Parameters
        ----------
        recordings : iterable
            Iterable of recording objects (RecordingBase or its subclasses).
        path : str, default="."
            Directory where recorded data will be saved.
        on_start : iterable, optional
            Iterable of objects with a `.start()` or `.run()` method. These
            methods are called when `Record.start()` is invoked. Note: `.start()`
            and `.run()` must be non-blocking.
        dt_request : float, default=0.7
            Time interval (in seconds) for checking user requests, such as
            graph pop-ups.
        """
        self.recordings = {rec.name: rec for rec in recordings}
        self.create_events()

        self.path = Path(path)
        self.path.mkdir(exist_ok=True)

        self.on_start = on_start

        self.dt_request = dt_request

        # Any additional functions that need to be run along the other threads
        # (to be defined in subclasses)
        self.additional_threads = []

    def __repr__(self):
        return f"Record object ({self.__class__.__name__})"

    # =========== Optional methods and attributes for subclassing ============

    def _save_metadata(self):
        """Save experimental metadata ; define in subclasses"""
        pass

    def save_metadata(self):
        """Save experimental metadata.

        self._save_metadata() needs to be defined in subclasses"""
        try:
            self._save_metadata()
        except Exception as e:
            # Since save_metadata is in a thread, it does not stop the main
            # program when an exception is thrown. As a result, the line
            # belows forces the program to stop when metadata saving fails.
            print("\n-----------------------------------")
            print(f"ERROR in metadata saving:\n{e}")
            print("CLI still running but PROGRAM STOPPED")
            print("--> PRESS Q TO EXIT")
            print("-----------------------------------\n")
            self.stop()

    def data_plot(self):
        """What to do with data when graph event is triggered"""
        pass

    # ============================= INIT METHODS =============================

    def create_events(self):
        """Create event objects managed by the CLI"""
        self.internal_stop = Event()  # event set to stop recording when needed.
        self.graph_request = (
            Event()
        )  # event set to start plotting the data in real time

        graph_event = ControlledEvent(
            event=self.graph_request,
            readable="graph",
            commands=("g", "graph"),
        )

        stop_event = ControlledEvent(
            event=self.internal_stop,
            readable="stop",
            commands=("q", "Q", "quit"),
        )

        self.controlled_events = graph_event, stop_event

    def parse_initial_user_commands(self, ppty_kwargs):
        """Check if user input contains specific properties.

        The values to apply for these properties are stored in a dict and
        will be applied to each recording when launched.

        If generic input (e.g. 'dt=10'), set all recordings to that value
        If specific input (e.g. dt_P=10), update recording to that value
        """
        initial_ppty_settings = {name: {} for name in self.recordings}

        global_commands = {}
        specific_commands = {}

        for cmd, value in ppty_kwargs.items():
            try:
                ppty_cmd, name = cmd.split("_", maxsplit=1)  # e.g. dt_P = 10
            except ValueError:  # e.g. dt=10
                global_commands[cmd] = value
            else:
                specific_commands[name] = ppty_cmd, value

        # Apply first global values to all recordings ------------------------
        for ppty_cmd, value in global_commands.items():
            for name in self.recordings:
                initial_ppty_settings[name][ppty_cmd] = value

        # Then apply commands to specific recordings if specified ------------
        for name, (ppty_cmd, value) in specific_commands.items():
            initial_ppty_settings[name][ppty_cmd] = value

        return initial_ppty_settings

    # =================== START RECORDING (MULTITHREAD) ======================

    def start_supplied_objects(self):
        """Start user-supplied non-blocking objects with a start() or run() method."""
        for obj in self.on_start:
            for method in "run", "start":
                try:
                    getattr(obj, method)()
                except AttributeError:
                    pass
                else:
                    break
            else:
                print(
                    f"WARNING - {obj.__class__.__name} does not have a"
                    f"run() or start() method --> not started. "
                )

    def start(self, **ppty_kwargs):
        """Start the recording process.

        Parameters
        ----------
        **ppty_kwargs : dict, optional
            Optional initial settings for properties. For example, use `dt=10` to
            change all time intervals to 10, or `dt_P=60` to change only the time
            interval of recording 'P'.
        """
        print(f"Recording started in folder {self.path.resolve()}")

        error_occurred = False
        self.threads = []

        try:
            # Check if user inputs particular initial settings for recordings
            init_ppties = self.parse_initial_user_commands(ppty_kwargs)

            # Start each recording with these initial settings
            for name, recording in self.recordings.items():
                recording.start(**init_ppties[name])

            # This is because in some situations save_metadata can take some
            # time, e.g. because it waits for sensors to be fully instantiated
            # to get info from them.
            metadata_thread = Thread(target=self.save_metadata)
            metadata_thread.start()
            self.threads.append(metadata_thread)

            # Add any other user-supplied functions for threading
            for func in self.additional_threads:
                thread = Thread(target=func)
                thread.start()
                self.threads.append(thread)

            # Add CLI. This one is a bit particular because it is blocking
            # with input() and has to be manually stopped. ----------------
            self.cli_thread = Thread(target=self.cli)
            self.cli_thread.start()

            # Start any other user-supplied objects (need to be nonblocking)
            self.start_supplied_objects()

            # real time graph (triggered by CLI, runs in main thread due to
            # GUI backend problems if not) --------------------------------
            self.data_graph()

        except Exception:
            error_occurred = True
            print("\nERROR during asynchronous Record. \n Stopping ... \n")
            print_exc()

        except KeyboardInterrupt:
            error_occurred = True
            print("\nManual interrupt by Ctrl-C event.\n Stopping ...\n")

        finally:
            self.stop()
            for recording in self.recordings.values():
                recording.stop()
                for thread in recording.threads:
                    thread.join()

            for thread in self.threads:
                thread.join()

            for obj in self.on_start:
                obj.stop()

            try:
                self.cli_thread
            except AttributeError:  # CLI thread not defined
                pass
            else:
                if error_occurred:
                    print('\nIMPORTANT: CLI still running. Input "q" to stop.\n')
                self.cli_thread.join()

            print("Recording Stopped")

    def stop(self):
        self.internal_stop.set()

    # =============================== Threads ================================

    @try_func
    def cli(self):
        if self.recordings:  # if no recordings provided, no need to record.
            command_input = CommandLineInterface(
                self.recordings,
                self.controlled_events,
            )
            command_input.run()
        else:
            self.internal_stop.set()
            raise ValueError("No recordings provided. Stopping ...")

    @try_func
    def data_graph(self):
        """Manage requests of real-time plotting of data during recording."""

        while not self.internal_stop.is_set():
            if self.graph_request.is_set():
                for recording in self.recordings.values():
                    recording.activate_queue("plotting")

                self.data_plot()  # Blocking (defined in subclasses)

                self.graph_request.clear()
                for recording in self.recordings.values():
                    recording.deactivate_queue("plotting")

            self.internal_stop.wait(self.dt_request)

    # ============================ Factory method ============================

    @classmethod
    def create(
        cls,
        mode=None,
        sensors=(),
        default_names=None,
        recording_types=None,
        recording_kwargs=None,
        programs=None,
        control_params=None,
        path=".",
        **kwargs,
    ):
        """Factory method to generate a Record object from sensor names and types.

        Parameters
        ----------
        mode : str or None, optional
            Sensor names in any order (e.g., 'PTB1') and potentially with separators
            (e.g., 'P-T-B1'). If None, all sensors are used by default.
        sensors : iterable, optional
            Iterable of sensor classes (e.g., [Sensor1, Sensor2]).
        default_names : iterable or None, optional
            Default sensor names to record if `mode` is not supplied.
        recording_types : dict, optional
            Dictionary mapping sensor names to their corresponding Recording class
            (RecordingBase or subclass) to be instantiated.
        recording_kwargs : dict, optional
            Dictionary mapping sensor names to their initialization keyword
            arguments.
        programs : dict, optional
            Dictionary mapping `mode` to an iterable of Program objects (or
            subclasses) from the `prevo.control.Program` class. The `mode` key
            describes all sensor recordings affected by the supplied programs.
        control_params : dict, optional
            Dictionary of `{command: kwargs}` to pass to program controls (e.g.,
            `dt`, `range_limits`). If None, default parameters are used (see
            `RecordingControl` class).
        path : str, default="."
            Directory where recorded data will be saved.
        **kwargs : dict, optional
            Additional keyword arguments to pass to the `Record.__init__` method.
            Possible options: `on_start` and `dt_request`.
        """
        all_sensors = {Sensor.name: Sensor for Sensor in sensors}
        name_info = {"possible_names": all_sensors, "default_names": default_names}
        names = mode_to_names(mode=mode, **name_info)

        programs = {} if programs is None else programs

        all_programs = {name: () for name in names}
        for pgm_mode, programs in programs.items():
            pgm_names = mode_to_names(mode=pgm_mode, **name_info)
            for name in pgm_names:
                all_programs[name] += programs

        recordings = []
        for name in names:
            Recording = recording_types[name]
            recording = Recording(
                Sensor=all_sensors[name],
                path=path,
                programs=all_programs[name],
                control_params=control_params,
                **recording_kwargs[name],
            )
            recordings.append(recording)

        return cls(recordings=recordings, path=path, **kwargs)



"""General file input/output with csv files"""

# ----------------------------- License information --------------------------

# This file is part of the prevo python package.
# Copyright (C) 2022 Olivier Vincent

# The prevo package is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# The prevo package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with the prevo python package.
# If not, see <https://www.gnu.org/licenses/>


# Standard library
from pathlib import Path

# Nonstandard
try:
    import pandas as pd
except ModuleNotFoundError:
    pass


class CsvFile:

    def __init__(
        self,
        path,
        column_names=None,
        column_formats=None,
        csv_separator='\t',
    ):
        """Init CsvFile object

        Parameters
        ----------

        path : {str, pathlib.Path}
            path to file, including filename and extension

        column_names : {array_like[str], None}, optional
            (optional, for saving data): iterable of column names

        column_formats : {array_like[str], None}, optional
            iterable of str formatings of data in columns

        path : {str, pathlib.Path}, optional
            folder in which file is located (default current folder)

        csv_separator: str
            separator used to separate data in file
        """
        self.path = Path(path)
        self.csv_separator = csv_separator
        self.column_names = column_names
        self.column_formats = column_formats

        if column_formats is None and self.column_names is not None:
            self.column_formats = ('',) * len(column_names)

    def load(self, nrange=None):
        """Load data recorded in path, possibly with a range of indices (n1, n2).

        Parameters
        ----------
        nrange : {tuple[int], None}
            select part of the data:
            - if nrange is None (default), load the whole file.
            - if nrange = (n1, n2), loads the file from line n1 to line n2,
              both n1 and n2 being included (first line of data is n=1).

        Returns
        -------
        pandas.DataFrame
            Pandas DataFrame of the requested size.
        """
        if nrange is None:
            kwargs = {}
        else:
            n1, n2 = nrange
            kwargs = {'skiprows': range(1, n1), 'nrows': n2 - n1 + 1}
        return pd.read_csv(self.path, delimiter=self.csv_separator, **kwargs)

    def number_of_lines(self):
        """Return number of lines of a file"""
        with open(self.path, 'r') as f:
            for i, line in enumerate(f):
                pass
            try:
                return i + 1
            except UnboundLocalError:  # handles the case of an empty file
                return 0

    def number_of_measurements(self):
        """Can be subclassed (here, assumes column titles)"""
        return self.number_of_lines() - 1

    # ---------- Methods that work on already opened file managers -----------

    def _init_file(self, file):
        """What to do with file when recording is started."""
        # Line below allows the user to re-start the recording and append data
        if self.number_of_lines() == 0:
            self._write_columns(file)

    def _write_columns(self, file):
        """How to init the file containing the data (when file already open)"""
        if self.column_names is None:
            return
        columns_str = f'{self.csv_separator.join(self.column_names)}\n'
        file.write(columns_str)

    def _write_line(self, data, file):
        """Save data to file when file is already open."""
        data_str = [f'{x:{fmt}}' for x, fmt in zip(data, self.column_formats)]
        line_for_saving = self.csv_separator.join(data_str) + '\n'
        file.write(line_for_saving)

    # ----------- Corresponding methods that open the file manager -----------

    def init_file(self):
        """What to do with file when recording is started."""
        with open(self.path, 'a', encoding='utf8') as file:
            self._init_file(file)


def resample_dataframe(df, rule, dropna=True):
    """Resample pandas dataframe containing a 'time (unix)' column.

    All other columns will be averaged.

    Parameters
    ----------
    df : pandas.DataFrame
        the data to resample

    rule : str
        pandas rule for resampling, e.g. "10s" for every 10 seconds

    dropna : bool
        if True (default), remove NaN values, e.g. if there is a time when
        there is no data e.g. because there was no data around the required
        time in the original file.
        if False, keep NaN values in order to have a constantly spaced
        time in the file.
    """
    df['datetime'] = pd.to_datetime(df['time (unix)'], unit='s')
    resampled_df = df.resample(rule=rule, on='datetime').mean()
    df.drop('datetime', axis=1, inplace=True)  # in order to keep original df
    new_df = resampled_df.reset_index().drop('datetime', axis=1)
    return new_df.dropna() if dropna else new_df


def resample_csv(
    file,
    rule,
    new_file=None,
    sep='\t',
    column_formats=None,
    dropna=True,
):
    """Resample data that has a 'time (unix)' column stored in csv file

    Takes data from the file and saves it into a new file.

    Parameters
    ----------
    file : {str, pathlib.Path}
        path to the file containing the data ; needs 'time (unix)' column

    rule : str
        pandas rule for resampling, e.g. "10s" for every 10 seconds

    new_file : {str, pathlib.Path}, optional
        if not supplied, use same name but with '_resampled' added in the name

    sep : str, optional
        separator used in the csv_file

    column_formats : array_like[str], optional
        iterable of f-string formatting of every column (including unix time)
        e.g. ('.3f', '.6f', '.0f')
        if None, use .3f for every column

    dropna : bool
        if True (default), remove NaN values, e.g. if there is a time when
        there is no data e.g. because there was no data around the required
        time in the original file.
        if False, keep NaN values in order to have a constantly spaced
        time in the file.
    """
    file = Path(file)
    data = pd.read_csv(file, sep=sep)
    new_data = resample_dataframe(data, rule=rule)

    if column_formats is not None:
        for name, fmt in zip(new_data, column_formats):
            new_data[name] = new_data[name].map(lambda x: f"{x:{fmt}}")
        float_format = None
    else:
        float_format = "%.3f"

    if new_file is None:
        new_file = file.with_name(f"{file.stem}_resampled{file.suffix}")

    new_data.to_csv(new_file, sep=sep, index=False, float_format=float_format)



"""Record several sensors as a function of time with interactive CLI and graph."""

# ----------------------------- License information --------------------------

# This file is part of the prevo python package.
# Copyright (C) 2022 Olivier Vincent

# The prevo package is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# The prevo package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with the prevo python package.
# If not, see <https://www.gnu.org/licenses/>

# Non-standard
import gittools
import prevo

# Local imports
from .general import RecordingBase, Record
from ..csv import CsvFile
from ..viewers import CvWindow, CvViewer
from ..viewers import TkWindow, TkViewer
from ..viewers import MplWindow, MplViewer
from ..misc import increment_filename

# Optional, nonstandard
try:
    from PIL import Image
except ModuleNotFoundError:
    pass


Windows = {
    "cv": CvWindow,
    "mpl": MplWindow,
    "tk": TkWindow,
}

Viewers = {
    "cv": CvViewer,
    "mpl": MplViewer,
    "tk": TkViewer,
}


class ImageRecording(RecordingBase):
    """Recording class for saving images and their associated timestamps."""

    def __init__(
        self,
        Sensor,
        timestamp_filename,
        path=".",
        image_path=None,
        extension=None,
        ndigits=5,
        quality=None,
        column_names=None,
        column_formats=None,
        csv_separator="\t",
        **kwargs,
    ):
        """Initialize an ImageRecording object.

        Parameters
        ----------
        Sensor : subclass of SensorBase
            Sensor class responsible for capturing images.
        timestamp_filename : str
            Name of the CSV/TSV file where timestamp data will be saved.
        path : str, default="."
            Directory where the CSV/TSV file will be created.
        image_path : str, optional
            Path where individual images will be saved. If None, a folder named
            after the sensor will be created in the `path` directory.
        extension : str, optional
            File extension for saved images (e.g., '.tif', '.jpg'). If None,
            a default extension will be used.
        ndigits : int, default=5
            Number of digits for the image counter in the filename.
        quality : int, optional
            Quality parameter for compressed image formats (e.g., JPEG, TIFF).
            Refer to the Pillow documentation for details:
            https://pillow.readthedocs.io/en/stable/handbook/image-file-formats.html
        column_names : iterable of str, optional
            Names of the columns in the CSV file.
        column_formats : iterable of str, optional
            Optional string formatting for each column in the CSV file.
        csv_separator : str, default="\t"
            Character used to separate columns in the CSV file.

        **kwargs : dict, optional
            Additional keyword arguments inherited from RecordingBase:
            dt : float, default=1
                Time interval (in seconds) between sensor readings.
            ctrl_ppties : iterable, optional
                Iterable of ControlledProperty objects to control during
                recording, in addition to default properties.
            active : bool, default=True
                If False, data reading is disabled until `self.active` is set
                to True.
            saving : bool, default=True
                If False, data will not be saved until `self.saving` is set to
                True. Data acquired while `saving=False` will not be saved later.
            continuous : bool, default=False
                If True, data is acquired as fast as possible from the sensor.
            warnings : bool, default=False
                If True, print warnings from the Timer (e.g., loop too short).
            precise : bool, default=False
                If True, use a precise timer in `oclock` (see `oclock.Timer`).
            immediate : bool, default=True
                If False, changes to the timer occur at the next timestep. If
                True, a new data point is taken immediately.
            programs : iterable, optional
                Iterable of Program objects for predefined temporal patterns of
                property changes.
            control_params : dict, optional
                Dictionary of `{command: kwargs}` to pass to program controls.
                If None, default parameters are used (see `RecordingControl`).
            dt_save : float, default=1.3
                Time interval (in seconds) for writing queues to files.
            dt_check : float, default=0.9
                Time interval (in seconds) for checking queue sizes.
        """
        super().__init__(Sensor=Sensor, path=path, **kwargs)

        # Here, file manager manages only the timestamp file, not the images
        self.file_manager = CsvFile(
            path=self.path / timestamp_filename,
            column_names=column_names,
            column_formats=column_formats,
            csv_separator=csv_separator,
        )

        self.image_path = self.path / Sensor.name if image_path is None else image_path
        self.image_path.mkdir(exist_ok=True)

        self.column_names = column_names

        if extension is None:
            # because for fast recording, .tif saving is much faster than png
            self.extension = ".tif" if self.continuous else ".png"
        else:
            self.extension = extension

        self.quality = quality
        self.ndigits = ndigits
        self.fmt = f"0{self.ndigits}"

        # number of images already recorded when record is called
        # (e.g. due to previous recording interrupted and restared)
        # The with open creates the file if not exists yet.
        with open(self.file_manager.path, "a", encoding="utf8"):
            n_lines = self.file_manager.number_of_lines()
        self.num = n_lines - 1 if n_lines > 1 else 0

    def format_measurement(self, data):
        """How to format the data"""
        return {"name": self.name, "num": self.num, **data}

    def after_measurement(self):
        """What to do after formatting data."""
        self.num += 1  # update number of recorded images for specific sensor

    def _generate_image_filename(self, measurement):
        """How to name images. Can be subclassed."""
        basename = f"{self.name}-{measurement['num']:{self.fmt}}"
        return basename + self.extension

    def _save_image(self, measurement, file):
        """How to save images to individual files. Can be subclassed."""
        img = measurement["image"]
        if self.quality is None:
            Image.fromarray(img).save(file)
        else:
            Image.fromarray(img).save(file, quality=self.quality)

    def _save_timestamp(self, measurement, file):
        """How to save timestamps and other info to (opened) timestamp file"""
        filename = self._generate_image_filename(measurement)
        info = {"filename": filename, **measurement}
        data = [info[x] for x in self.column_names]
        self.file_manager._write_line(data, file)

    def save(self, measurement, file):
        """Write data to .tsv file with format: datetime / delta_t / value(s).

        Input
        -----
        - data: dict of data from the read() function
        - file: file in which to save the data
        """
        img_filename = self._generate_image_filename(measurement)
        self._save_image(measurement, file=self.image_path / img_filename)
        self._save_timestamp(measurement, file=file)

    @property
    def info(self):
        """Additional metadata info to save in metadata file. Subclass."""
        return {}


class ImageRecord(Record):
    """Main class for managing simultaneous temporal recordings of images."""

    def __init__(
        self,
        recordings,
        metadata_filename="Images_Metadata.json",
        checked_modules=(),
        dt_graph=0.02,
        viewer="tk",
        dirty_ok=True,
        **kwargs,
    ):
        """Initialize an ImageRecord object.

        Parameters
        ----------
        recordings : iterable
            Iterable of recording objects (RecordingBase or its subclasses).
        metadata_filename : str, default="Images_Metadata.json"
            Name of the JSON file where metadata will be saved.
        checked_modules : iterable, optional
            Iterable of Python modules whose versions will be saved in the
            metadata, in addition to the `prevo` module.
        dt_graph : float, default=0.02
            Time interval (in seconds) to refresh the numerical graph.
        viewer : str, default="tk"
            Type of viewer to display images. Possible values: 'mpl', 'tk', 'cv'.
        dirty_ok : bool, default=True
            If False, the recording cannot be started if Git repositories are
            not clean (uncommitted changes).

        **kwargs : dict, optional
            Additional keyword arguments inherited from Record:
            path : str, default="."
                Directory where recorded data will be saved.
            on_start : iterable, optional
                Iterable of objects with `.start()` or `.run()` methods. These
                methods are called when `Record.start()` is invoked. Note:
                `.start()` and `.run()` must be non-blocking.
            dt_request : float, default=0.7
                Time interval (in seconds) for checking user requests, such as
                graph pop-ups.
        """

        super().__init__(recordings=recordings, **kwargs)
        self.metadata_filename = metadata_filename
        self.checked_modules = set((prevo,) + tuple(checked_modules))

        # Viewing options -------------
        self.dt_graph = dt_graph
        self.viewer = viewer

        self.dirty_ok = dirty_ok
        self.get_image_recordings()

    def get_image_recordings(self):
        """Useful when combined with other recording types (e.g. vacuum)"""
        self.image_recordings = {}
        for name, recording in self.recordings.items():
            if issubclass(recording.__class__, ImageRecording):
                self.image_recordings[name] = recording

    def _save_metadata(self):
        """To be able to call save_metadata() with arbitrary filenames"""
        metadata_file = self.path / self.metadata_filename

        if metadata_file.exists():
            metadata_file = increment_filename(metadata_file)

        info = {}

        for name, recording in self.image_recordings.items():
            info[name] = {
                "sensor": repr(recording.Sensor),
                "initial image number": recording.num,
                "extension": recording.extension,
                "digit number": recording.ndigits,
                "quality": str(recording.quality),
                **recording.info,
            }

        gittools.save_metadata(
            metadata_file,
            info=info,
            module=self.checked_modules,
            dirty_warning=True,
            dirty_ok=self.dirty_ok,
            notag_warning=True,
            nogit_ok=True,
            nogit_warning=True,
        )

    def data_plot(self):
        """What to do when graph event is triggered"""

        Viewer = Viewers[self.viewer]
        Window = Windows[self.viewer]

        windows = []
        for name, recording in self.image_recordings.items():
            image_queue = recording.queues["plotting"]
            win = Window(
                image_queue,
                name=name,
                show_num=True,
                show_fps=True,
            )
            windows.append(win)

        viewer = Viewer(
            windows,
            external_stop=self.internal_stop,
            dt_graph=self.dt_graph,
        )
        viewer.start()



"""Viewers for live series of images (e.g. from cameras) arriving as queues."""


# ----------------------------- License information --------------------------

# This file is part of the prevo python package.
# Copyright (C) 2022 Olivier Vincent

# The prevo package is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# The prevo package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with the prevo python package.
# If not, see <https://www.gnu.org/licenses/>


import tkinter as tk
import itertools
from PIL import Image, ImageTk

from .general import WindowBase, ViewerBase, CONFIG, DISPOSITIONS


class TkWindow(WindowBase):
    """Live view of images using tkinter"""

    def __init__(
        self,
        image_queue,
        auto_size=True,
        **kwargs,
    ):
        """Init TkSingleViewer object

        Parameters
        ----------

        - image_queue: queue in which taken images are put.
        - auto_size: autoscale image to window in real time

        Additional kwargs from WindowBase:
        - name: optional name for display purposes.
        - calculate_fps: if True, store image times fo calculate fps
        - show_fps: if True, indicate current display fps on viewer
        - show_num: if True, indicate current image number on viewer
                    (note: image data must be a dict with key 'num', or
                    a different data_formatter must be provided)
        - dt_fps: how often (in seconds) display fps are calculated
        - dt_num: how often (in seconds) image numbers are updated
        - measurement_formatter: object that transforms elements from the
                                 queue into image arrays and image numbers
                                 (type MeasurementFormatter or equivalent)
        """
        super().__init__(image_queue, **kwargs)
        self.auto_size = auto_size

    @property
    def parent(self):
        """Tkinter parent in which to put the current window."""
        return self._parent

    @parent.setter
    def parent(self, value):
        self._parent = value
        self._parent.configure(bg=CONFIG['bgcolor'])

    def _init_window(self):
        """Create tkinter window and elements."""
        if self.name is not None:
            self.title_label = tk.Label(
                self.parent,
                text=self.name,
                font=(CONFIG['fontfamily'], 14),
                bg=CONFIG['bgcolor'],
                fg=CONFIG['textcolor'],
            )
            self.title_label.pack(expand=True)

        self.image_label = tk.Label(self.parent, highlightthickness=0)
        self.image_label.pack(expand=True)

        if self.info_queues:
            self.info_label = tk.Label(
                self.parent,
                bg=CONFIG['bgcolor'],
                fg=CONFIG['textcolor'],
                font=(CONFIG['fontfamily'], 12),
                text=str('...'),
            )
            self.info_label.pack(expand=True)

        self.image_count = 0

    def _display_info(self):
        self.info_label.config(text=self.info)

    def _display_image(self):
        """How to display image in viewer."""
        self.image_count += 1

        img = Image.fromarray(self.image)
        img_disp = self._prepare_displayed_image(img)

        self.img = ImageTk.PhotoImage(image=img_disp)
        self.image_label.configure(image=self.img)

    def _prepare_displayed_image(self, img):
        """Resize image and/or calculate aspect ratio if necessary"""

        if self.image_count > 1:
            if self.auto_size:
                dimensions = self._adapt_image_to_window()
                try:
                    img_disp = img.resize(dimensions, Image.LANCZOS)
                except ValueError:  # somtimes dimensions are (0, 0) for some reason
                    img_disp = img
            else:
                img_disp = img

        else:  # Calculate aspect ratio on first image received
            self.aspect_ratio = img.height / img.width
            img_disp = img

        return img_disp

    def _adapt_image_to_window(self):
        """Calculate new dimensions of image to accommodate window resizing."""

        window_width = self.parent.winfo_width()
        window_height = self.parent.winfo_height()

        target_width = 0.98 * window_width
        target_height = 0.85 * window_height

        target_ratio = target_height / target_width

        if target_ratio > self.aspect_ratio:
            width = int(target_width)
            height = int(target_width * self.aspect_ratio)
        else:
            height = int(target_height)
            width = int(target_height / self.aspect_ratio)

        return width, height


class TkViewer(ViewerBase):
    """Live view of images from multiple cameras using tkinter"""

    def __init__(
        self,
        windows,
        fit_to_screen=True,
        **kwargs,
    ):
        """Init TkViewer object

        Parameters
        ----------

        - windows: iterable of objects of type WindowBase or subclasses
        - fit_to_screen: maximize window size when instantiated
        - root: Tkinter parent in which to display viewer (if not, tk.Tk())

        Additional kwargs from ViewerBase
        - external_stop: stopping event (threading.Event or equivalent)
                         signaling stopping requested from outside of the class
                         (won't be set or cleared, just monitored)
        - dt_graph: how often (in seconds) the viewer is updated
        """
        self.fit_to_screen = fit_to_screen
        super().__init__(windows=windows, **kwargs)

    def _create_root(self):
        self.root = tk.Tk()
        self.root.configure(bg=CONFIG['bgcolor'])
        # Detect manual closing of window
        self.root.protocol("WM_DELETE_WINDOW", self._on_close)

    def _create_grid(self, parent=None):
        """This allows to create the elements in other parents for child classes."""
        parent = self.root if parent is None else parent

        for window in self.windows:
            window.parent = tk.Frame(master=parent)

        n = len(self.windows)
        n1, n2 = DISPOSITIONS[n]  # dimensions of grid to place elements
        positions = itertools.product(range(n1), range(n2))

        for window, position in zip(self.windows, positions):
            i, j = position
            window.parent.grid(row=i, column=j, padx=5, pady=5, sticky='nsew')

        # Make columns and rows expand and be all the same size
        # Note: the str in uniform= is just an identifier
        # all columns / rows sharing the same string are kept of same size
        for i in range(n1):
            parent.grid_rowconfigure(i, weight=1, uniform='same rows')
        for j in range(n2):
            parent.grid_columnconfigure(j, weight=1, uniform='same columns')

    def _init_viewer(self):
        self._create_root()
        self._create_grid(parent=self.root)
        if self.fit_to_screen:
            self._fit_to_screen()

    def _fit_to_screen(self):
        """Adapt window size to screen resolution/size"""
        w_screen = self.root.winfo_screenwidth()
        h_screen = self.root.winfo_screenheight()
        self.root.geometry(f"{0.8 * w_screen:.0f}x{0.8 * h_screen:.0f}")

    def _run(self):
        self._update()
        self.root.mainloop()

    def _update(self):
        self._update_info()
        self._update_images()
        self._check_external_stop()
        self.loop = self.root.after(int(1000 * self.dt_graph), self._update)

    def _cancel_loop(self):
        try:
            loop = self.loop
        except AttributeError:  # in case after() has not been called yet
            pass
        else:
            self.root.after_cancel(loop)

    def _on_close(self):
        """Callback to user manually closing window"""
        self._cancel_loop()
        self.root.destroy()  # without this the window doesn't get closed

    def stop(self):
        self._cancel_loop()
        # If application still active, close tk window
        try:
            self.root.destroy()
        except tk.TclError:
            pass
        super().stop()



"""General tools and base classes for the prevo.plot module."""

# ----------------------------- License information --------------------------

# This file is part of the prevo python package.
# Copyright (C) 2022 Olivier Vincent

# The prevo package is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# The prevo package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with the prevo python package.
# If not, see <https://www.gnu.org/licenses/>

# Standard library imports
from abc import ABC, abstractmethod
from datetime import datetime
from threading import Event

# Non standard imports
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
from matplotlib import cm
import tzlocal

# Local imports
from ..misc import get_last_from_queue

# The two lines below have been added following a console FutureWarning:
# "Using an implicitly registered datetime converter for a matplotlib plotting
# method. The converter was registered by pandas on import. Future versions of
# pandas will require you to explicitly register matplotlib converters."
try:
    import pandas as pd
    from pandas.plotting import register_matplotlib_converters

    register_matplotlib_converters()
except ModuleNotFoundError:
    pandas_available = False
else:
    pandas_available = True


# How to place elements on window as a function of number of widgets
DISPOSITIONS = {
    1: (1, 1),
    2: (1, 2),
    3: (1, 3),
    4: (2, 2),
}

# Misc =======================================================================

local_timezone = tzlocal.get_localzone()


class MeasurementFormatter:
    """Format lists, arrays etc. for plotting in matplotlib.

    Can be subclassed or replaced to adapt to applications."""

    @staticmethod
    def format_measurement(measurement):
        """Transform measurement from the queue into something usable by manage_data()

        Can be subclassed to adapt to various applications.
        Here, assumes data is incoming in the form of a dictionary with at
        least keys:
        - 'name' (str, identifier of sensor)
        - 'time (unix)' (floar or array of floats)
        - 'values' (iterable of values, or iterable of arrays of values)

        Subclass possible to adapt to applications.
        """
        data = {key: measurement[key] for key in ("name", "values", "time (unix)")}
        return data

    # ==== Methods to transform lists into arrays for matplotlib plotting ====

    @staticmethod
    def list_of_single_values_to_array(datalist):
        """How to convert list of single values to a numpy array."""
        return datalist

    @staticmethod
    def list_of_single_times_to_array(timelist):
        """How to convert list of single times to a numpy array."""
        return timelist

    @staticmethod
    def list_of_value_arrays_to_array(datalist):
        """How to convert list of arrays of values to a numpy array."""
        return np.concatenate(datalist)

    @staticmethod
    def list_of_time_arrays_to_array(timelist):
        """How to convert list array of times to a numpy array."""
        return np.concatenate(timelist)

    # ============= Methods to convert unix times into datetimes =============

    @staticmethod
    def to_datetime_datetime(unix_time):
        """Transform single value of unix time into timezone-aware datetime."""
        return datetime.fromtimestamp(unix_time, local_timezone)

    @staticmethod
    def to_datetime_numpy(unix_times):
        """Transform iterable / array of unix times into datetimes.

        Note: this is the fastest method, but the datetimes are in UTC format
              (not local time)
        """
        return (np.array(unix_times) * 1e9).astype("datetime64[ns]")

    @staticmethod
    def to_datetime_pandas(unix_times):
        """Transform iterable / array of datetimes into pandas Series.

        Note: here, the datetimes are in local timezone format, but this is
              slower than the numpy approach.
        """
        # For some reason, it's faster (and more precise) to convert to numpy first
        np_times = (np.array(unix_times) * 1e9).astype("datetime64[ns]")
        pd_times = pd.Series(np_times)
        return pd.to_datetime(pd_times, utc=True).dt.tz_convert(local_timezone)


class GraphBase(ABC):
    """Base class for managing plotting of arbitrary measurement data"""

    def __init__(
        self,
        names,
        data_types,
        fig=None,
        colors=None,
        legends=None,
        linestyles=None,
        linestyle=".",
        data_as_array=False,
        time_conversion="numpy",
        measurement_formatter=MeasurementFormatter(),
    ):
        """Initialize figures and axes for plotting data as a function of specified types.

        Parameters
        ----------
        names : iterable
            Names of recordings/sensors to be plotted.
        data_types : dict
            Dictionary with recording names as keys and corresponding data types as
            values. Can contain more keys than those in `names`.
        fig : matplotlib.figure.Figure, optional
            Matplotlib figure in which to draw the graph.
        colors : dict, optional
            Dictionary of colors with keys 'fig', 'ax', and recording names.
        legends : dict, optional
            Dictionary of legend names (iterable) for all channels of each sensor,
            keyed by recording names.
        linestyles : dict, optional
            Dictionary of linestyles (iterable) to distinguish channels and sensors,
            keyed by recording names. If None, all lines use the `linestyle` parameter.
            If only some recordings are specified, others use the default or
            `linestyle`.
        linestyle : str, default='.'
            Matplotlib linestyle (e.g., '.', '-', '.-').
        data_as_array : bool or dict, default=False
            If sensors return arrays of values for different times, set to True. Can also
            be a dictionary of booleans keyed by recording names if some sensors return
            arrays and others do not.
        time_conversion : str, default='numpy'
            Method to convert Unix time to datetime for arrays. Possible values:
            'numpy', 'pandas'.
        measurement_formatter : MeasurementFormatter, optional
            MeasurementFormatter (or subclass) object.
        """

        self.names = names
        self.data_types = {name: data_types[name] for name in self.names}
        self.fig = fig
        self.colors = colors
        self.legends = legends if legends is not None else {}
        self.linestyles = linestyles if linestyles is not None else {}
        self.linestyle = linestyle

        self.current_data = self.create_empty_data()

        self.measurement_formatter = measurement_formatter
        self.manage_array_conversion(data_as_array=data_as_array)
        self.manage_time_conversion(time_conversion=time_conversion)

        self.create_axes()
        self.set_colors()
        self.format_graph()
        self.create_lines()
        self.fig.tight_layout()

        # Create onclick callback to activate / deactivate autoscaling
        self.cid = self.fig.canvas.mpl_connect("button_press_event", self.onclick)

    # ========================== Methods to subclass =========================

    @abstractmethod
    def create_axes(self):
        """To be defined in subclasses. Returns fig, axs"""
        pass

    @abstractmethod
    def format_graph(self):
        """To be defined in subclasses.

        Misc. settings for graph (time formatting, limits etc.)"""
        pass

    @abstractmethod
    def update_data(self, data):
        """Store measurement time and values in active data lists."""
        pass

    @abstractmethod
    def update(self):
        """How to update graph after adding measurements to it (graph.add())."""
        pass

    @property
    def animated_artists(self):
        """Optional property to define for graphs updated with blitting."""
        return ()

    # =========================== Static Plotting methods ===========================

    def add(self, measurement):
        """Add measurement on graph, possibly adding to existing ones."""
        # The line below allows some sensors to avoid being plotted by reading
        # None when called, and also to not consider empty queues.
        if measurement is not None:
            data = self.measurement_formatter.format_measurement(measurement)
            self.update_data(data)

    # ===================== Graph initialization methods =====================

    @property
    def all_data_types(self):
        """Return a set of all datatypes corresponding to the active names."""
        all_types = ()
        for name in self.names:
            data_types = self.data_types[name]
            all_types += data_types
        return set(all_types)

    def set_colors(self):
        """ "Define fig/ax colors if supplied"""
        if self.colors is None:
            self.colors = {}
        else:
            figcolor = self.colors.get("fig", "white")
            self.fig.set_facecolor(figcolor)
            for ax in self.axs.values():
                axcolor = self.colors.get("ax", "white")
                ax.set_facecolor(axcolor)
                ax.grid()

        missing_color_names = []
        n_missing_colors = 0
        for name, dtypes in self.data_types.items():
            try:
                self.colors[name]
            except (KeyError, TypeError):
                missing_color_names.append(name)
                n_missing_colors += len(dtypes)

        if not n_missing_colors:
            return

        m = cm.get_cmap("tab10", n_missing_colors)
        i = 0
        for name in missing_color_names:
            dtypes = self.data_types[name]
            colors = []
            for _ in dtypes:
                colors.append(m.colors[i])
                i += 1
            self.colors[name] = tuple(colors)

    def create_lines(self):
        """Create lines for each value of each sensor"""
        self.lines = {}
        self.lines_list = []

        for name in self.names:
            dtypes = self.data_types[name]
            clrs = self.colors[name]
            labels = self.legends.get(name, [None] * len(dtypes))
            lstyles = self.linestyles.get(name, [self.linestyle] * len(dtypes))

            self.lines[name] = []

            for dtype, clr, label, lstyle in zip(dtypes, clrs, labels, lstyles):
                # Plot data in correct axis depending on type
                ax = self.axs[dtype]
                (line,) = ax.plot([], [], lstyle, color=clr, label=label)

                self.lines[name].append(line)
                # Below, used for returning animated artists for blitting
                self.lines_list.append(line)

        if self.legends:
            legend_clr = self.colors.get("legend")
            for ax in self.axs.values():
                ax.legend(loc="lower left", facecolor=legend_clr)

    def create_empty_data(self):
        data = {}
        for name in self.names:
            times = []
            values = []
            for _ in self.data_types[name]:
                values.append([])
            data[name] = {"times": times, "values": values}
        return data

    @staticmethod
    def onclick(event):
        """Activate/deactivate autoscale by clicking to allow for data inspection.

        - Left click (e.g. when zooming, panning, etc.): deactivate autoscale
        - Right click: reactivate autoscale.
        """
        ax = event.inaxes
        if ax is None:
            pass
        elif event.button == 1:  # left click
            ax.axes.autoscale(False, axis="both")
        elif event.button == 3:  # right click
            ax.axes.autoscale(True, axis="both")
        else:
            pass

    # ========================== Conversion methods ==========================

    def manage_array_conversion(self, data_as_array):
        """Determine conversion methods from data to arrays"""
        try:
            data_as_array.get  # no error if dict
        except AttributeError:  # it's a bool: put info for all sensors
            self.data_as_array = {name: data_as_array for name in self.names}
        else:
            self.data_as_array = data_as_array

        self.datalist_to_array = {}
        self.timelist_to_array = {}

        for name, data_as_array in self.data_as_array.items():
            if data_as_array:
                self.datalist_to_array[name] = (
                    self.measurement_formatter.list_of_value_arrays_to_array
                )
                self.timelist_to_array[name] = (
                    self.measurement_formatter.list_of_time_arrays_to_array
                )
            else:
                self.datalist_to_array[name] = (
                    self.measurement_formatter.list_of_single_values_to_array
                )
                self.timelist_to_array[name] = (
                    self.measurement_formatter.list_of_single_times_to_array
                )

    def manage_time_conversion(self, time_conversion):
        """How to convert input times into datetimes manageable for plotting."""

        time_converters = {
            "datetime": self.measurement_formatter.to_datetime_datetime,
            "numpy": self.measurement_formatter.to_datetime_numpy,
        }
        
        if pandas_available:
            time_converters["pandas"] = self.measurement_formatter.to_datetime_pandas

        self.time_converters = {}
        for name in self.names:
            if self.data_as_array[name]:
                self.time_converters[name] = time_converters[time_conversion]
            else:
                self.time_converters[name] = time_converters["datetime"]

    # ================= Update graph with data from queue(s) =================

    def run(
        self,
        queues,
        external_stop=None,
        dt_graph=0.1,
        blit=False,
    ):
        """Run a live view of the plot using data from queues.

        Convenience method to instantiate an UpdateGraph object.

        Parameters
        ----------
        queues : iterable
            Iterable of queues to read data from.
        external_stop : threading.Event, optional
            External stop request. If set, closes the figure.
        dt_graph : float, default=0.1
            Time interval (in seconds) to update the graph.
        blit : bool, default=False
            If True, use blitting to speed up the Matplotlib animation.
        """
        update_graph = UpdateGraph(
            graph=self,
            queues=queues,
            external_stop=external_stop,
            dt_graph=dt_graph,
            blit=blit,
        )
        update_graph.run()

    def close(self):
        """Close matplotlib figure associated with graph"""
        plt.close(self.fig)


class UpdateGraph:
    def __init__(
        self,
        graph,
        queues,
        external_stop=None,
        dt_graph=0.1,
        blit=False,
    ):
        """Update the plot with data received from a queue.

        Parameters
        ----------
        graph : GraphBase or subclass
            Object of the GraphBase class or its subclasses.
        queues : iterable
            Iterable of queues to read data from.
        external_stop : threading.Event or equivalent, optional
            Stopping event signaling a stop request from outside the class.
            This event is only monitored and not modified.
        dt_graph : float, default=0.1
            Time interval (in seconds) to update the graph.
        blit : bool, default=False
            If True, use blitting to speed up the Matplotlib animation.
        """
        self.graph = graph
        self.queues = queues
        self.dt_graph = dt_graph
        self.blit = blit

        self.external_stop = external_stop
        self.internal_stop = Event()

        self.graph.fig.canvas.mpl_connect("close_event", self.on_fig_close)

    def on_fig_close(self, event):
        """What to do when figure is closed."""
        self.stop()

    def plot_new_data(self, i=0):
        """define what to do at each loop of the matplotlib animation."""

        if self.internal_stop.is_set():
            return

        if self.external_stop is not None and self.external_stop.is_set():
            self.stop()
            self.graph.close()

        for queue in self.queues:
            measurement = get_last_from_queue(queue)
            self.graph.add(measurement)

        self.graph.update()

        if self.blit:
            return self.graph.animated_artists

    def run(self):
        # Below, it doesn't work if there is no ani = before the FuncAnimation
        ani = FuncAnimation(
            fig=self.graph.fig,
            func=self.plot_new_data,
            interval=self.dt_graph * 1000,
            cache_frame_data=False,
            save_count=0,
            blit=self.blit,
        )

        plt.show(block=True)  # block=True allows the animation to work even
        # when matplotlib is in interactive mode (plt.ion()).

        return ani

    def stop(self):
        self.internal_stop.set()



"""Viewers for live series of images (e.g. from cameras) arriving as queues."""


# ----------------------------- License information --------------------------

# This file is part of the prevo python package.
# Copyright (C) 2022 Olivier Vincent

# The prevo package is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# The prevo package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with the prevo python package.
# If not, see <https://www.gnu.org/licenses/>


import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation

from .general import max_possible_pixel_value
from .general import WindowBase, ViewerBase, CONFIG, DISPOSITIONS


class MplWindow(WindowBase):
    """Display camera images using Matplotlib"""

    def __init__(
        self,
        image_queue,
        **kwargs,
    ):
        """Init MplSingleViewer object.

        Parameters
        ----------

        - image_queue: queue in which taken images are put.
        _ blit: if True, use blitting for faster rendering (can cause issues
                for updating info such as fps, image number)

        Additional kwargs from WindowBase
        - image_queue: queue in which taken images are put.
        - name: optional name for display purposes.
        - calculate_fps: if True, store image times fo calculate fps
        - show_fps: if True, indicate current display fps on viewer
        - show_num: if True, indicate current image number on viewer
                    (note: image data must be a dict with key 'num', or
                    a different data_formatter must be provided)
        - dt_fps: how often (in seconds) display fps are calculated
        - dt_num: how often (in seconds) image numbers are updated
        - measurement_formatter: object that transforms elements from the
                                 queue into image arrays and image numbers
                                 (type MeasurementFormatter or equivalent)

        NOTE: self.ax must be defined before calling various window methods.
        """
        super().__init__(image_queue, **kwargs)

    @property
    def ax(self):
        """Matplotlib axes object in which to put the current window."""
        return self._ax

    @ax.setter
    def ax(self, value):
        self._ax = value

    def _init_window(self):
        self._format_axes()
        self.init_done = False

    def _format_axes(self):
        """"Set colors, title etc."""
        self.ax.set_title(
            self.name,
            color=CONFIG['textcolor'],
            fontfamily=CONFIG['fontfamily'],
        )

        for location in 'bottom', 'top', 'left', 'right':
            # self.ax.spines[location].set_color(textcolor)
            self.ax.spines[location].set_visible(False)

        self.ax.xaxis.label.set_color(CONFIG['textcolor'])
        self.ax.tick_params(axis='both', colors=CONFIG['textcolor'])

    def _init_image(self, image):
        kwargs = {} if image.ndim > 2 else {'cmap': 'gray'}

        self.im = self.ax.imshow(
            image,
            animated=True,
            vmin=0,
            vmax=max_possible_pixel_value(image),
            **kwargs,
        )
        self.init_done = True
        self.xlabel = self.ax.set_xlabel(
            '...',
            color=CONFIG['textcolor'],
            fontfamily=CONFIG['fontfamily'],
        )

    def _update(self, i=0):
        """Indicate what happens at each step of the matplotlib animation."""
        self._update_info()
        data = self._update_image()
        return () if data is None else (self.im,)

    def _display_info(self):
        try:
            self.xlabel.set_text(self.info)
        # In case the _init_image has not been called yet
        except AttributeError:
            pass

    def _display_image(self):
        """How to display image in viewer."""
        if not self.init_done:
            self._init_image(self.image)
        else:
            self.im.set_array(self.image)


class MplViewer(ViewerBase):
    """Display several cameras at the same time using Matplotlib"""

    def __init__(
        self,
        windows,
        fig=None,
        blit=True,
        **kwargs,
    ):
        """
        Init MplViewer object

        Parameters
        ----------

        - windows: iterable of objects of type WindowBase or subclasses
        - fig: (optional): matplotlib figure in which to create the viewer.
        - blit: if True, use blitting for faster rendering (can cause issues
                for updating info such as fps, image number)


        Additional kwargs from ViewerBase
        - external_stop: stopping event (threading.Event or equivalent)
                         signaling stopping requested from outside of the class
                         (won't be set or cleared, just monitored)
        - dt_graph: how often (in seconds) the viewer is updated
        """
        self.blit = blit
        self.fig = fig
        super().__init__(windows=windows, **kwargs)

    def _init_viewer(self):
        """Generate figure/axes as a function of input names"""

        n = len(self.windows)
        n1, n2 = DISPOSITIONS[n]  # dimensions of grid to place elements

        if self.fig is None:
            width = 4 * n2
            height = 4 * n1
            self.fig = plt.figure(figsize=(width, height))

        for i, window in enumerate(self.windows):
            ax = self.fig.add_subplot(n1, n2, i + 1)
            window.ax = ax

        self.fig.set_facecolor(CONFIG['bgcolor'])
        self.fig.tight_layout()

        # NOTE: stop() is called automatically by start() once matplotlib
        # window is destroyed. But if one wants more things to happen
        # upon figure closing, one can do something like:
        # self.fig.canvas.mpl_connect('close_event', self._on_close)

        # NOTE: I also tried to play with self.ani.event_source.stop()
        # to prevent the bug in tkinter on_timer, but no success.

    def _update(self, i=0):
        """Indicate what happens at each step of the matplotlib animation."""
        self._check_external_stop()
        to_be_animated = ()
        for window in self.windows:
            to_be_animated += window._update(i=i)
        return to_be_animated

    def _run(self):
        """Main function to run the animation"""
        self.ani = FuncAnimation(
            self.fig,
            self._update,
            interval=int(self.dt_graph) * 1000,
            blit=self.blit,
            cache_frame_data=False,
        )
        plt.show(block=True)
        return self.ani

    def stop(self):
        plt.close(self.fig)
        super().stop()



"""Plot data from sensors in oscilloscope-like fashion"""

# ----------------------------- License information --------------------------

# This file is part of the prevo python package.
# Copyright (C) 2022 Olivier Vincent

# The prevo package is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# The prevo package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with the prevo python package.
# If not, see <https://www.gnu.org/licenses/>

# Standard library imports
import time

# Non standard imports
import numpy as np
import matplotlib.pyplot as plt

# Local imports
from .general import GraphBase, MeasurementFormatter


class OscilloMeasurementFormatter(MeasurementFormatter):
    """Overwrite some formatting methods from the default."""

    def list_of_single_values_to_array(self, datalist):
        """To be able to filter on condition and concatenate"""
        return np.array(datalist, dtype=np.float64)

    def list_of_single_times_to_array(self, timelist):
        """To be able to filter on condition and concatenate"""
        return np.array(timelist, dtype=np.float64)


class OscilloGraph(GraphBase):
    def __init__(
        self,
        names,
        data_types,
        data_ranges,
        window_width=10,
        fig=None,
        colors=None,
        legends=None,
        linestyles=None,
        linestyle=".",
        data_as_array=False,
        measurement_formatter=OscilloMeasurementFormatter(),
    ):
        """Initialize figures and axes for plotting data as a function of specified types.

        Parameters
        ----------
        names : iterable
            Names of recordings/sensors to be plotted.
        data_types : dict
            Dictionary with recording names as keys and corresponding data types
            as values. Can contain more keys than those in `names`.
        data_ranges : dict
            Dictionary with possible data types as keys and corresponding value
            ranges as values. Used to set initial y-axis limits. Can contain more
            keys than actual data types used.
        window_width : float
            Width (in seconds) of the displayed window.
        fig : matplotlib.figure.Figure, optional
            Matplotlib figure in which to draw the graph.
        colors : dict, optional
            Dictionary of colors with keys 'fig', 'ax', 'bar', and recording
            names.
        legends : dict, optional
            Dictionary of legend names (iterable) for all channels of each
            sensor, keyed by recording names.
        linestyles : dict, optional
            Dictionary of linestyles (iterable) to distinguish channels and
            sensors, keyed by recording names. If None, all lines use the
            `linestyle` parameter. If only some recordings are specified, others
            use the default or `linestyle`.
        linestyle : str, default='.'
            Matplotlib linestyle (e.g., '.', '-', '.-').
        data_as_array : bool or dict, default=False
            If sensors return arrays of values for different times, set to True.
            Can also be a dictionary of booleans keyed by recording names if
            some sensors return arrays and others do not.
        measurement_formatter : MeasurementFormatter, optional
            MeasurementFormatter (or subclass) object.
        """
        self.data_ranges = data_ranges
        self.window_width = window_width
        self.reference_time = None

        super().__init__(
            names=names,
            data_types=data_types,
            fig=fig,
            colors=colors,
            legends=legends,
            linestyles=linestyles,
            linestyle=linestyle,
            data_as_array=data_as_array,
            measurement_formatter=measurement_formatter,
        )

        self.create_bars()
        self.previous_data = (
            self.create_empty_data()
        )  # current_data created by the base class

    # ================== Methods subclassed from GraphBase ===================

    def create_axes(self):
        """Generate figure/axes as a function of input data types"""

        n = len(self.all_data_types)
        self.fig, axes = plt.subplots(n, 1)

        # Transform axes into a tuple if only one ax
        try:
            iter(axes)
        except TypeError:
            axes = (axes,)

        self.axs = {}
        for ax, datatype in zip(axes, self.all_data_types):
            ax.set_ylabel(datatype)
            self.axs[datatype] = ax

    def format_graph(self):
        """Misc. settings for graph (time formatting, limits etc.)"""

        w = self.window_width
        for dtype, ax in self.axs.items():
            ax.set_xlim((-0.001 * w, 1.001 * w))
            ax.set_ylim(self.data_ranges[dtype])
            ax.grid()

    def update_data(self, data):
        tmin, tmax = self.get_time_boundaries(data)

        if self.reference_time is None:
            self.reference_time = tmin  # Take time of 1st data as time 0

        self.update_stored_data(
            data=data,
            stored_data=self.current_data,
        )

        # In case measurement arrives late after window has already refreshed,
        # duplicate it to previous data so that it is remains visible
        # This is particularly useful for data arriving as arrays; in which
        # case the array will be duplicated to appear both at the beginning
        # and end of the window when the times in the arrays span values
        # across the window wrapping time.
        if tmin < self.reference_time:
            self.update_stored_data(
                data=data,
                stored_data=self.previous_data,
            )

        # There is no need to do the same for 'future' points that would arrive
        # with tmax > reference_time + window_size, because in
        # principle all data arriving is from the past or present.

    def update(self):
        self.update_lines()
        self.update_bars()
        if self.reference_time and (self.relative_time > self.window_width):
            self.wrap()

    @property
    def animated_artists(self):
        artists = self.lines_list + list(self.bars.values())
        return artists

    # ======================= Other graph init methods =======================

    def create_bars(self):
        """Create traveling bars"""
        self.bars = {}
        for dtype, ax in self.axs.items():
            barcolor = self.colors.get("bar", "silver")
            bar = ax.axvline(0, linestyle="-", c=barcolor, linewidth=4)
            self.bars[dtype] = bar

    # =============== Methods overriden from the parent class ================

    def on_click():
        """Here turn off autoscale, which can cause problems with blitting."""
        pass

    # ======= Methods that can be subclassed to adapt to applications ========

    def get_time_boundaries(self, data):
        """Subclass if necessary."""
        t = data["time (unix)"]
        name = data["name"]

        if self.data_as_array[name]:
            return t[0], t[-1]
        else:
            return t, t

    # ========================== Misc. properties ============================

    @property
    def current_time(self):
        return time.time()

    @property
    def relative_time(self):
        return self.current_time - self.reference_time

    # ========================= Graph Update Methods =========================

    def wrap(self):
        """What to do each time the bars exceeed window size"""
        self.previous_data = self.current_data
        self.current_data = self.create_empty_data()
        self.reference_time += self.window_width

    def update_stored_data(self, data, stored_data):
        """Store measurement time and values in active data lists.

        Parameters
        ----------
        data: data as output by format_measurement()
        stored_data: either self.current_data or self.previous_data.
        """
        name = data["name"]
        time = data["time (unix)"]
        values = data["values"]

        stored_data[name]["times"].append(time)
        for i, value in enumerate(values):
            stored_data[name]["values"][i].append(value)

    def update_lines(self):
        """Update line positions with current data."""

        for name in self.lines:
            lines = self.lines[name]
            previous_data = self.previous_data[name]
            current_data = self.current_data[name]

            rel_times = []

            # Avoids problems if no data stored yet
            prev_exists = bool(previous_data["times"])
            curr_exists = bool(current_data["times"])

            if not (prev_exists or curr_exists):
                continue

            if curr_exists:
                curr_times = self.timelist_to_array[name](current_data["times"])
                curr_rel_times = curr_times - self.reference_time
                rel_times.append(curr_rel_times)

            if prev_exists:
                prev_times = self.timelist_to_array[name](previous_data["times"])
                prev_condition = prev_times + self.window_width > self.current_time
                prev_rel_times = (
                    prev_times[prev_condition] - self.reference_time + self.window_width
                )
                rel_times.append(prev_rel_times)

            rel_times_array = np.concatenate(rel_times)

            for line, prev_values, curr_values in zip(
                lines,
                previous_data["values"],
                current_data["values"],
            ):
                vals = []

                if curr_exists:
                    curr_vals = self.datalist_to_array[name](curr_values)
                    vals.append(curr_vals)

                if prev_exists:
                    prev_vals = self.datalist_to_array[name](prev_values)
                    vals.append(prev_vals[prev_condition])

                values_array = np.concatenate(vals)

                line.set_data(rel_times_array, values_array)

    def update_bars(self):
        if self.reference_time:  # Avoids problems if no data arrived yet
            t = self.relative_time
            for bar in self.bars.values():
                bar.set_xdata(t)



"""Record several sensors as a function of time with interactive CLI and graph."""

# ----------------------------- License information --------------------------

# This file is part of the prevo python package.
# Copyright (C) 2022 Olivier Vincent

# The prevo package is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# The prevo package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with the prevo python package.
# If not, see <https://www.gnu.org/licenses/>

# Non-standard imports
import gittools
import prevo

# Local imports
from .general import RecordingBase, Record
from ..csv import CsvFile
from ..plot import NumericalGraph
from ..misc import increment_filename


class NumericalRecording(RecordingBase):
    """Recording class for saving numerical sensor data to CSV files."""

    def __init__(
        self,
        Sensor,
        filename,
        path=".",
        column_names=None,
        column_formats=None,
        csv_separator="\t",
        **kwargs,
    ):
        """Initialize a NumericalRecording object.

        Parameters
        ----------
        Sensor : subclass of SensorBase
            Sensor class responsible for capturing numerical data.
        filename : str
            Name of the CSV/TSV file where data will be saved.
        path : str, default="."
            Directory where the CSV/TSV file will be created.
        column_names : iterable of str, optional
            Names of the columns in the CSV file.
        column_formats : iterable of str, optional
            Optional string formatting for each column in the CSV file.
        csv_separator : str, default="\t"
            Character used to separate columns in the CSV
        """
        super().__init__(Sensor=Sensor, path=path, **kwargs)

        self.file_manager = CsvFile(
            path=self.path / filename,
            column_names=column_names,
            column_formats=column_formats,
            csv_separator=csv_separator,
        )

    def format_measurement(self, measurement):
        """Format raw sensor data into a measurement object.

        This object is put in the saving queue and in the plotting queue
        if active.

        Here we assume that measurements are in the form of dictionaries, but
        this can be changed by subclassing.
        """

        if measurement is None:
            return
        # Adding the name is useful e.g. for plotting.
        measurement["name"] = self.name
        return measurement

    def save(self, measurement, file):
        """Save measurement data to a file.

        Parameters
        ----------
        measurement : Measurement object
            Measurement object to be saved. Assumed to be a dictionary with keys:
            'time (unix)', 'dt (s)', and 'values'.
        file : file-like object
            File object where the data will be saved.

        Returns
        -------
        iterable
            Iterable of data to be saved in the CSV file. The length of the iterable
            must match the length of `column_names`.

        Notes
        -----
        - Can be redefined in subclasses for custom behavior.
        - If `measurement` is None, `Record.data_save()` will not save the data.
        """
        time_info = (measurement["time (unix)"], measurement["dt (s)"])
        value_info = measurement["values"]
        self.file_manager._write_line(time_info + value_info, file=file)


class NumericalRecord(Record):
    """Class for managing simultaneous temporal recordings of numerical sensors."""

    def __init__(
        self,
        recordings,
        metadata_filename="Numerical_Metadata.json",
        checked_modules=(),
        data_types=None,
        dt_graph=0.1,
        graph_legends=None,
        graph_colors=None,
        graph_linestyle=".",
        dirty_ok=True,
        **kwargs,
    ):
        """Initialize a NumericalRecord object.

        Parameters
        ----------
        recordings : iterable
            Iterable of recording objects (RecordingBase or its subclasses).
        metadata_filename : str, default="Numerical_Metadata.json"
            Name of the JSON file where metadata will be saved.
        checked_modules : iterable, optional
            Iterable of Python modules whose versions will be saved in the
            metadata, in addition to the `prevo` module.
        data_types : iterable, optional
            Iterable of data types for selecting the window in which to plot
            data when graphs are active. If not supplied, graphs cannot be
            instantiated.
        dt_graph : float, default=0.1
            Time interval (in seconds) to refresh the numerical graph.
        graph_colors : dict, optional
            Dictionary of graph colors for the numerical graph (see `prevo.plot`).
        graph_legends : dict, optional
            Dictionary of graph legends for the numerical graph (see `prevo.plot`).
        graph_linestyle : str, default="."
            Linestyle of data on the numerical graph (e.g., '.-').
        dirty_ok : bool, default=True
            If False, the recording cannot be started if Git repositories are
            not clean (uncommitted changes).

        **kwargs : dict, optional
            Additional keyword arguments inherited from Record:
            path : str, default="."
                Directory where recorded data will be saved.
            on_start : iterable, optional
                Iterable of objects with `.start()` or `.run()` methods. These
                methods are called when `Record.start()` is invoked. Note:
                `.start()` and `.run()` must be non-blocking.
            dt_request : float, default=0.7
                Time interval (in seconds) for checking user requests, such as
                graph pop-ups.
        """
        super().__init__(recordings=recordings, **kwargs)
        self.metadata_filename = metadata_filename
        self.checked_modules = set((prevo,) + tuple(checked_modules))

        # Graphing options -------------
        self.data_types = data_types
        self.dt_graph = dt_graph
        self.graph_legends = graph_legends
        self.graph_colors = graph_colors
        self.graph_linestyle = graph_linestyle

        self.dirty_ok = dirty_ok
        self.get_numerical_recordings()

    def get_numerical_recordings(self):
        """Useful when vacuum combined with other recording types (e.g. camrec)"""
        self.numerical_recordings = {}
        for name, recording in self.recordings.items():
            if issubclass(recording.__class__, NumericalRecording):
                self.numerical_recordings[name] = recording

    def _save_metadata(self):
        """To call save_metadata() with custom filenames"""
        metadata_file = self.path / self.metadata_filename

        if metadata_file.exists():
            metadata_file = increment_filename(metadata_file)

        gittools.save_metadata(
            metadata_file,
            module=self.checked_modules,
            dirty_warning=True,
            dirty_ok=self.dirty_ok,
            notag_warning=True,
            nogit_ok=True,
            nogit_warning=True,
        )

    def data_plot(self):
        """What to do when graph event is triggered"""
        if self.data_types is None:
            print("WARNING --- No data types supplied. Graph not possible.")
            return

        graph = NumericalGraph(
            names=self.numerical_recordings,
            data_types=self.data_types,
            legends=self.graph_legends,
            colors=self.graph_colors,
            linestyle=self.graph_linestyle,
            data_as_array=False,
        )

        # In case the queue contains other measurements than numerical
        # (e.g. images from cameras)
        numerical_queues = [
            recording.queues["plotting"]
            for recording in self.numerical_recordings.values()
        ]

        graph.run(
            queues=numerical_queues,
            external_stop=self.internal_stop,
            dt_graph=self.dt_graph,
        )



"""Viewers for live series of images (e.g. from cameras) arriving as queues."""


# ----------------------------- License information --------------------------

# This file is part of the prevo python package.
# Copyright (C) 2022 Olivier Vincent

# The prevo package is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# The prevo package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with the prevo python package.
# If not, see <https://www.gnu.org/licenses/>


from abc import ABC, abstractmethod
import time
from queue import Queue
from threading import Thread, Event
from traceback import print_exc

import numpy as np

from ..misc import get_all_from_queue, get_last_from_queue


# ========================== Appearance Parameters  ==========================


CONFIG = {
    'bgcolor': '#485a6a',
    'textcolor': '#e7eff6',
    'fontfamily': "serif",
}


# How to place elements on window as a function of number of widgets
DISPOSITIONS = {
    1: (1, 1),
    2: (1, 2),
    3: (1, 3),
    4: (2, 2),
}


# =============================== MISC. Tools ================================


def max_possible_pixel_value(img):
    """Return max pixel value depending on image type, for use in plt.imshow.

    Input
    -----
    img: numpy array

    Output
    ------
    vmax: max pixel value (int or float or None)
    """
    if img.dtype == 'uint8':
        return 2**8 - 1
    elif img.dtype == 'uint16':
        return 2**16 - 1
    else:
        return None


class InfoSender(ABC):
    """Class to send information to display in Image Viewer.

    For example: fps, image info, etc.
    """
    def __init__(
        self,
        queue=None,
        dt_check=1,
    ):
        """Init info sender object

        Parameters
        ----------

        - queue: queue into information is put
        - dt_check: how often (in seconds) information is sent
        """
        self.queue = Queue() if queue is None else queue
        self.internal_stop = Event()
        self.dt_check = dt_check

    @abstractmethod
    def _generate_info(self):
        """To be defined in subclass.

        Should return a str of info to print in the viewer.
        Should return a false-like value if no news info can be provided."""
        pass

    def _run(self):
        """Send information periodically (blocking)."""

        while not self.internal_stop.is_set():
            info = self._generate_info()
            self.queue.put(info)

            # dt_check should be long compared to the time required to
            # process and generate the info.
            self.internal_stop.wait(self.dt_check)

    def start(self):
        """Same as _run() but nonblocking."""
        Thread(target=self._run).start()

    def stop(self):
        self.internal_stop.set()


class LiveFpsCalculator(InfoSender):
    """"Calculate fps in real time from a queue supplying image times.

    fps values are sent back in another queue as str
    """

    def __init__(
        self,
        time_queue,
        queue=None,
        dt_check=1,
    ):
        """Parameters:

        - time_queue: queue from which display times arrive
        - queue: queue into which fps values are put
        - dt_check: how often (in seconds) times are checked to calculate fps
        """
        super().__init__(queue=queue, dt_check=dt_check)
        self.time_queue = time_queue

    def _generate_info(self):
        """Calculate fps from display times, print '...' if none available"""
        times = get_all_from_queue(self.time_queue)
        if times:
            if len(times) > 1:
                fps = 1 / np.diff(times).mean()
                return f'[{fps:.1f} fps]'
        return '[... fps]'


class LiveImageNumber(InfoSender):
    """"Calculate fps in real time from a queue supplying image times.

    fps values are sent back in another queue as str
    """

    def __init__(
        self,
        num_queue,
        queue=None,
        dt_check=1,
    ):
        """Parameters:

        - num_queue: queue from which image numbers arrive
        - queue: queue into which output values are put
        - dt_check: how often (in seconds) times are checked to calculate fps
        """
        super().__init__(queue=queue, dt_check=dt_check)
        self.num_queue = num_queue
        self.last_num = None

    def _generate_info(self):
        """Return last image number received from queue."""
        num = get_last_from_queue(self.num_queue)
        if num is not None:
            self.last_num = num

        if self.last_num is None:
            return '[# ...]'
        else:
            return f'[# {self.last_num}]'


class MeasurementFormatter:
    """How to transform elements from the queue into image arrays and img number.

    Can be subclassed.
    """

    def get_image(self, measurement):
        """How to transform individual elements from the queue into an image.

        (returns an array-like object).
        Can be subclassed to accommodate different queue formats.
        """
        return measurement['image']

    def get_num(self, measurement):
        """How to get image numbers from individual elements from the queue.

        (returns an int).
        Can be subclassed to accommodate different queue formats.
        """
        return measurement['num']


default_measurement_formatter = MeasurementFormatter()


# =============================== Base classes ===============================

class WindowBase:
    """Base class for windows managing single image queues."""

    def __init__(
        self,
        image_queue,
        name=None,
        calculate_fps=False,
        show_fps=False,
        show_num=False,
        dt_fps=2,
        dt_num=0.2,
        measurement_formatter=default_measurement_formatter,
    ):
        """Init Window object.

        Parameters
        ----------

        - image_queue: queue in which taken images are put.
        - name: optional name for display purposes.
        - calculate_fps: if True, store image times fo calculate fps
        - show_fps: if True, indicate current display fps on viewer
        - show_num: if True, indicate current image number on viewer
                    (note: image data must be a dict with key 'num', or
                    a different data_formatter must be provided)
        - dt_fps: how often (in seconds) display fps are calculated
        - dt_num: how often (in seconds) image numbers are updated
        - measurement_formatter: object that transforms elements from the
                                 queue into image arrays and image numbers
                                 (type MeasurementFormatter or equivalent)
        """
        self.image_queue = image_queue
        self.name = name

        self.calculate_fps = calculate_fps
        self.show_fps = show_fps
        self.show_num = show_num

        self.measurement_formatter = measurement_formatter
        self.internal_stop = Event()

        try:
            self._init_info(dt_fps=dt_fps, dt_num=dt_num)
        except Exception:
            print(f'--- !!! Error in  {self.name} Window Init !!! ---')
            print_exc()
            self.stop()

    def __repr__(self):
        return f'{self.__class__.__name__} ({self.name})'

    def _init_info(self, **kwargs):
        """Init info objects that manage printing of fps, img number etc."""

        self.info_senders = []
        self.info_queues = {}
        self.info_values = {}

        # store times at which images are shown on screen (e.g. for fps calc.)
        if self.calculate_fps:
            self.display_times = []             # to calculate fps on all times

        if self.show_fps:
            self.display_times_queue = Queue()  # to calculate fps on partial data
            fps_calculator = LiveFpsCalculator(time_queue=self.display_times_queue,
                                               dt_check=kwargs.get('dt_fps'))
            self.info_queues['fps'] = fps_calculator.queue
            self.info_values['fps'] = ''
            self.info_senders.append(fps_calculator)
            fps_calculator.start()

        if self.show_num:
            self.image_number_queue = Queue()
            image_number = LiveImageNumber(num_queue=self.image_number_queue,
                                           dt_check=kwargs.get('dt_num'))
            self.info_queues['num'] = image_number.queue
            self.info_values['num'] = ''
            self.info_senders.append(image_number)
            image_number.start()

    def _store_display_times(self):
        t = time.perf_counter()
        if self.calculate_fps:
            self.display_times.append(t)
        if self.show_fps:
            self.display_times_queue.put(t)

    def _init_window(self):
        """How to create/init window."""
        pass

    def _display_info(self):
        """How to display information from info queues on image.

        Define in subclasses.
        """
        pass

    def _display_image(self):
        """How to display image in viewer.

        Define in subclasses.
        """
        pass

    def _get_info(self):
        """Get information from info queues and display it if not None.

        Returns None if no new info to print on screen.
        """
        update = False
        for name, queue in self.info_queues.items():
            info = get_last_from_queue(queue)
            if info:
                update = True
                self.info_values[name] = info
        if update:
            new_info = ' '.join(self.info_values.values())
            return new_info

    def _update_info(self):
        """Typically one wants to call this regularly even if no new image is
        coming in the queue, because info of fps etc. is stored in queues and
        can have some delay depending on how quickly the queues are probed."""
        info = self._get_info()
        if info is not None:
            self.info = info
            self._display_info()

        # Can be useful if one must decide what to do depending on info
        # (e.g. if info is None, do nothing else)
        return info

    def _update_image(self):
        """How to process measurement from the image queue"""
        data = get_last_from_queue(self.image_queue)
        if data is not None:

            self.image = self.measurement_formatter.get_image(data)

            if self.show_num:
                num = self.measurement_formatter.get_num(data)
                self.image_number_queue.put(num)

            if self.calculate_fps or self.show_fps:
                self._store_display_times()

            self._display_image()

        # Can be useful if one must decide what to do depending on data
        # (e.g. if data is None, do nothing else)
        return data

    def stop(self):
        """What to do when live viewer is stopped"""
        self.internal_stop.set()

        # Stop threads that caculate fps, img number etc.
        for info_sender in self.info_senders:
            info_sender.stop()

        if self.calculate_fps:
            if len(self.display_times) > 1:
                fps = 1 / np.diff(self.display_times).mean()
                print(f'Average display frame rate [{self.name}]: {fps:.3f} fps. ')
            else:
                print('Impossible to calculate average FPS (not enough values). ')


class ViewerBase:
    """Base class for Viewers (contain windows)"""

    def __init__(
        self,
        windows,
        external_stop=None,
        dt_graph=0.02,
    ):
        """Init ViewerBase object

        Parameters
        ----------
        - windows: iterable of objects of type WindowBase or subclasses
        - external_stop: stopping event (threading.Event or equivalent)
                         signaling stopping requested from outside of the class
                         (won't be set or cleared, just monitored)
        - dt_graph: how often (in seconds) the viewer is updated
        """
        self.windows = windows
        self.dt_graph = dt_graph
        self.external_stop = external_stop
        self.internal_stop = Event()

    def _init_viewer(self):
        """Define in subclasses"""

    def _run(self):
        """Define in subclasses"""
        pass

    def _init_windows(self):
        for window in self.windows:
            window._init_window()

    def _update_info(self):
        for window in self.windows:
            window._update_info()

    def _update_images(self):
        for window in self.windows:
            window._update_image()

    def _check_external_stop(self):
        if self.external_stop:
            if self.external_stop.is_set():
                self.stop()

    def start(self):
        try:
            self._init_viewer()
            self._init_windows()
            self._run()
        except Exception:
            print('--- !!! Error in Viewer !!! ---')
            print_exc()
        self.stop()

    def stop(self):
        self.internal_stop.set()
        for window in self.windows:
            window.stop()



"""Plot data from sensors (from live measurements or saved data)."""

# ----------------------------- License information --------------------------

# This file is part of the prevo python package.
# Copyright (C) 2022 Olivier Vincent

# The prevo package is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# The prevo package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with the prevo python package.
# If not, see <https://www.gnu.org/licenses/>

import matplotlib.pyplot as plt
import matplotlib.dates as mdates

from .general import GraphBase, MeasurementFormatter
from .general import DISPOSITIONS, local_timezone


# =============================== Main classes ===============================


class NumericalGraph(GraphBase):
    def __init__(
        self,
        names,
        data_types,
        fig=None,
        colors=None,
        legends=None,
        linestyles=None,
        linestyle=".",
        data_as_array=False,
        time_conversion="numpy",
        measurement_formatter=MeasurementFormatter(),
    ):
        """Initialize figures and axes for plotting data as a function of types.

        Parameters
        ----------
        names : iterable
            Names of recordings/sensors to be plotted.
        data_types : dict
            Dictionary with recording names as keys and corresponding data types
            as values. Can contain more keys than those in `names`.
        fig : matplotlib.figure.Figure, optional
            Matplotlib figure in which to draw the graph.
        colors : dict, optional
            Dictionary of colors with keys 'fig', 'ax', and recording names.
        legends : dict, optional
            Dictionary of legend names (iterable) for all channels of each sensor,
            keyed by recording names.
        linestyles : dict, optional
            Dictionary of linestyles (iterable) to distinguish channels and
            sensors, keyed by recording names. If None, all lines use the
            `linestyle` parameter. If only some recordings are specified, others
            use the default or `linestyle`.
        linestyle : str, default='.'
            Matplotlib linestyle (e.g., '.', '-', '.-').
        data_as_array : bool, default=False
            If sensors return arrays of values for different times, set to True.
        time_conversion : str, default='numpy'
            Method to convert Unix time to datetime for arrays. Possible: 'numpy',
            'pandas'.
        measurement_formatter : MeasurementFormatter, optional
            MeasurementFormatter (or subclass) object.
        """
        super().__init__(
            names=names,
            data_types=data_types,
            fig=fig,
            colors=colors,
            legends=legends,
            linestyles=linestyles,
            linestyle=linestyle,
            data_as_array=data_as_array,
            time_conversion=time_conversion,
            measurement_formatter=measurement_formatter,
        )

    # ================== Methods subclassed from GraphBase ===================

    def create_axes(self):
        """Generate figure/axes as a function of input data types"""

        n = len(self.all_data_types)
        if n > 4:
            msg = f"Mode combination {self.all_data_types} not supported yet"
            raise Exception(msg)

        n1, n2 = DISPOSITIONS[n]  # dimensions of grid to place elements

        if self.fig is None:
            self.fig = plt.figure()

        width = 5 * n2
        height = 3 * n1
        self.fig.set_figheight(height)
        self.fig.set_figwidth(width)

        self.axs = {}
        for i, datatype in enumerate(self.all_data_types):
            ax = self.fig.add_subplot(n1, n2, i + 1)
            ax.set_ylabel(datatype)
            self.axs[datatype] = ax

    def format_graph(self):
        """Misc. settings for graph (time formatting, limits etc.)"""
        self.locators = {}  # For concise formatting of time -----------------
        self.formatters = {}
        for ax in self.axs.values():
            locator = mdates.AutoDateLocator(tz=local_timezone)
            self.locators[ax] = locator
            self.formatters[ax] = mdates.ConciseDateFormatter(locator, tz=local_timezone)

    def update_data(self, data):
        """Store measurement time and values in active data lists."""

        name = data["name"]
        values = data["values"]
        time = self.time_converters[name](data["time (unix)"])

        self.current_data[name]["times"].append(time)
        for i, value in enumerate(values):
            self.current_data[name]["values"][i].append(value)

    def update(self):
        self.update_lines()
        self.update_time_formatting()

    @property
    def animated_artists(self):
        return self.lines_list

    # ======================== Local update methods ==========================

    def update_lines(self):
        """Update line positions with current data."""

        for name in self.names:
            lines = self.lines[name]
            current_data = self.current_data[name]

            if current_data["times"]:  # Avoids problems if no data stored yet
                times = self.timelist_to_array[name](current_data["times"])

                for line, curr_values in zip(lines, current_data["values"]):
                    values = self.datalist_to_array[name](curr_values)
                    line.set_data(times, values)

    def update_time_formatting(self):
        """Use Concise Date Formatting for minimal space used on screen by time"""
        for ax in self.axs.values():
            ax.xaxis.set_major_locator(self.locators[ax])
            ax.xaxis.set_major_formatter(self.formatters[ax])
            # Lines below are needed for autoscaling to work
            ax.relim()
            ax.autoscale_view(tight=True, scalex=True, scaley=True)



"""Viewers for live series of images (e.g. from cameras) arriving as queues."""


# ----------------------------- License information --------------------------

# This file is part of the prevo python package.
# Copyright (C) 2022 Olivier Vincent

# The prevo package is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# The prevo package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with the prevo python package.
# If not, see <https://www.gnu.org/licenses/>


from .general import ViewerBase, WindowBase

try:
    import cv2
except ModuleNotFoundError:
    pass


class CvWindow(WindowBase):
    """Display camera images using OpenCV"""

    def __init__(
        self,
        image_queue,
        name,
        **kwargs,
    ):
        """Init Window object.

        Parameters
        ----------

        - image_queue: queue in which taken images are put.
        - name: NOT optional here because it serves as ID for openCV windows

        Additional kwargs from WindowBase:
        - calculate_fps: if True, store image times fo calculate fps
        - show_fps: if True, indicate current display fps on viewer
        - show_num: if True, indicate current image number on viewer
                    (note: image data must be a dict with key 'num', or
                    a different data_formatter must be provided)
        - dt_fps: how often (in seconds) display fps are calculated
        - dt_num: how often (in seconds) image numbers are updated
        - measurement_formatter: object that transforms elements from the
                                 queue into image arrays and image numbers
                                 (type MeasurementFormatter or equivalent)
        """
        super().__init__(image_queue, name=name, **kwargs)

    def _init_window(self):
        """Create window"""
        cv2.namedWindow(self.name, cv2.WINDOW_NORMAL)
        self.info = '...'

    def _display_image(self):
        # Here we need to have the info display directly in display_image
        # since the info is written directly on the image itself
        # This can cause some imprecisions in the image numbers displayed
        cv2.putText(self.image, self.info, (50, 50), cv2.FONT_HERSHEY_SIMPLEX,
                    1, (255, 255, 255), 2, cv2.LINE_AA)

        if self.image.ndim > 2:
            # openCV works with BGR data
            self.image = cv2.cvtColor(self.image, cv2.COLOR_RGB2BGR)
        cv2.imshow(self.name, self.image)


class CvViewer(ViewerBase):
    """Display several cameras at the same time using OpenCV"""

    def __init__(
        self,
        windows,
        **kwargs,
    ):
        """Init CvViewer object

        Parameters
        ----------

        - windows: iterable of objects of type WindowBase or subclasses

        Additional kwargs from ViewerBase
        - external_stop: stopping event (threading.Event or equivalent)
                         signaling stopping requested from outside of the class
                         (won't be set or cleared, just monitored)
        - dt_graph: how often (in seconds) the viewer is updated
        """
        super().__init__(windows=windows, **kwargs)

    def _run(self):
        """Loop to run live viewer"""
        open_windows = []
        for window in self.windows:
            wopen = cv2.getWindowProperty(window.name, cv2.WND_PROP_VISIBLE) > 0
            open_windows.append(wopen)

        while (any(open_windows)):
            self._update_info()
            self._update_images()
            self._check_external_stop()
            if self.internal_stop.is_set():
                for window in self.windows:
                    cv2.destroyWindow(window.name)
                break
            cv2.waitKey(int(self.dt_graph * 1000))

    def stop(self):
        super().stop()
        cv2.destroyAllWindows()



"""Tests for the prevo package.

(VERY partial)
"""

# Standard library
from pathlib import Path

# local imports
import prevo
from prevo.csv import CsvFile, resample_csv


DATAFOLDER = Path(prevo.__file__).parent / '..' / 'data'
DATA_FILE = DATAFOLDER / 'manip' / 'Vacuum_Pressure.tsv'
out_file_1 = DATAFOLDER / 'untracked_data' / 'Pressure_resampled_1.tsv'
out_file_2 = DATAFOLDER / 'untracked_data' / 'Pressure_resampled_2.tsv'


def test_resample_simple():
    """Resampling of numerical data, default parameters"""
    resample_csv(DATA_FILE, '30s', new_file=out_file_1)
    csv_file = CsvFile(out_file_1)
    assert csv_file.path.exists()
    assert csv_file.number_of_measurements() == 3758


def test_resample_advanced():
    """Resampling of numerical data"""

    resample_csv(
        DATA_FILE,
        '2min',
        column_formats=('.3f', '.3f', '.2f'),
        new_file=out_file_2,
    )
    csv_file = CsvFile(out_file_2)
    assert csv_file.path.exists()
    assert csv_file.number_of_measurements() == 940
